"""
Generic function parser for functions embedded in a string, on the form
`$funcname(*args, **kwargs)`, for example:

```
"A string $foo() with $bar(a, b, c, $moo(), d=23) etc."
```

Each arg/kwarg can also be another nested function. These will be executed
inside-out and their return will used as arguments for the enclosing function
(so the same as for regular Python function execution).

This is the base for all forms of embedded func-parsing, like inlinefuncs and
protfuncs. Each function available to use must be registered as a 'safe'
function for the parser to accept it. This is usually done in a module with
regular Python functions on the form:

```python
# in a module whose path is passed to the parser

def _helper(x):
    # use underscore to NOT make the function available as a callable

def funcname(*args, **kwargs):
    # this can be accessed as $funcname(*args, **kwargs)
    # it must always accept *args and **kwargs.
    ...
    return something
```

Usage:

```python
from evennia.utils.funcparser import FuncParser

parser = FuncParser("path.to.module_with_callables")
result = parser.parse("String with $funcname() in it")

```

The `FuncParser` also accepts a direct dict mapping of `{'name': callable, ...}`.

---

"""
import dataclasses
import inspect
import random

from django.conf import settings

from evennia.utils import logger, search
from evennia.utils.utils import (
    callables_from_module,
    crop,
    int2str,
    justify,
    make_iter,
    pad,
    safe_convert_to_types,
    variable_from_module,
)
from evennia.utils.verb_conjugation.conjugate import verb_actor_stance_components
from evennia.utils.verb_conjugation.pronouns import pronoun_to_viewpoints

# setup

_CLIENT_DEFAULT_WIDTH = settings.CLIENT_DEFAULT_WIDTH
_MAX_NESTING = settings.FUNCPARSER_MAX_NESTING
_START_CHAR = settings.FUNCPARSER_START_CHAR
_ESCAPE_CHAR = settings.FUNCPARSER_ESCAPE_CHAR


@dataclasses.dataclass
class _ParsedFunc:
    """
    Represents a function parsed from the string

    """

    prefix: str = _START_CHAR
    funcname: str = ""
    args: list = dataclasses.field(default_factory=list)
    kwargs: dict = dataclasses.field(default_factory=dict)

    # state storage
    fullstr: str = ""
    infuncstr: str = ""
    rawstr: str = ""
    double_quoted: int = -1
    current_kwarg: str = ""
    open_lparens: int = 0
    open_lsquate: int = 0
    open_lcurly: int = 0
    exec_return = ""

    def get(self):
        return self.funcname, self.args, self.kwargs

    def __str__(self):
        return self.prefix + self.rawstr + self.infuncstr


class ParsingError(RuntimeError):
    """
    Failed to parse for some reason.
    """

    pass


class FuncParser:
    """
    Sets up a parser for strings containing `$funcname(*args, **kwargs)`
    substrings.

    """

    def __init__(
        self,
        callables,
        start_char=_START_CHAR,
        escape_char=_ESCAPE_CHAR,
        max_nesting=_MAX_NESTING,
        **default_kwargs,
    ):
        """
        Initialize the parser.

        Args:
            callables (str, module, list or dict): Where to find
                'safe' functions to make available in the parser. If a `dict`,
                it should be a direct mapping `{"funcname": callable, ...}`. If
                one or mode modules or module-paths, the module(s) are first checked
                for a dict `FUNCPARSER_CALLABLES = {"funcname", callable, ...}`. If
                no such variable exists, all callables in the module (whose name does
                not start with an underscore) will be made available to the parser.
            start_char (str, optional): A character used to identify the beginning
                of a parseable function. Default is `$`.
            escape_char (str, optional): Prepend characters with this to have
                them not count as a function. Default is the backtick, `\\\\`.
            max_nesting (int, optional): How many levels of nested function calls
                are allowed, to avoid exploitation. Default is 20.
            **default_kwargs: These kwargs will be passed into all callables. These
                kwargs can be overridden both by kwargs passed direcetly to `.parse` *and*
                by kwargs given directly in the string `$funcname` call. They are
                suitable for global defaults that is intended to be changed by the
                user. To guarantee a call always gets a particular kwarg, pass it
                into `.parse` as `**reserved_kwargs` instead.

        """
        if isinstance(callables, dict):
            loaded_callables = {**callables}
        else:
            # load all modules/paths in sequence. Later-added will override
            # earlier same-named callables (allows for overriding evennia defaults)
            loaded_callables = {}
            for module_or_path in make_iter(callables):
                callables_mapping = variable_from_module(
                    module_or_path, variable="FUNCPARSER_CALLABLES"
                )
                if callables_mapping:
                    try:
                        # mapping supplied in variable
                        loaded_callables.update(callables_mapping)
                    except ValueError:
                        raise ParsingError(
                            f"Failure to parse - {module_or_path}.FUNCPARSER_CALLABLES "
                            "(must be a dict {'funcname': callable, ...})"
                        )
                else:
                    # use all top-level variables
                    # (handles both paths and module instances
                    loaded_callables.update(callables_from_module(module_or_path))
        self.validate_callables(loaded_callables)
        self.callables = loaded_callables
        self.escape_char = escape_char
        self.start_char = start_char
        self.default_kwargs = default_kwargs

    def validate_callables(self, callables):
        """
        Validate the loaded callables. Each callable must support at least
        `funcname(*args, **kwargs)`.
        property.

        Args:
            callables (dict): A mapping `{"funcname": callable, ...}` to validate

        Raise:
            AssertionError: If invalid callable was found.

        Notes:
            This is also a good method to override for individual parsers
            needing to run any particular pre-checks.

        """
        for funcname, clble in callables.items():
            try:
                mapping = inspect.getfullargspec(clble)
            except TypeError:
                logger.log_trace(f"Could not run getfullargspec on {funcname}: {clble}")
            else:
                assert mapping.varargs, f"Parse-func callable '{funcname}' does not support *args."
                assert mapping.varkw, f"Parse-func callable '{funcname}' does not support **kwargs."

    def execute(self, parsedfunc, raise_errors=False, **reserved_kwargs):
        """
        Execute a parsed function

        Args:
            parsedfunc (_ParsedFunc): This dataclass holds the parsed details
                of the function.
            raise_errors (bool, optional): Raise errors. Otherwise return the
                string with the function unparsed.
            **reserved_kwargs: These kwargs are _guaranteed_ to always be passed into
                the callable on every call. It will override any default kwargs
                _and_ also a same-named kwarg given manually in the $funcname
                call. This is often used by Evennia to pass required data into
                the callable, for example the current Session for inlinefuncs.
        Returns:
            any: The result of the execution. If this is a nested function, it
                can be anything, otherwise it will be converted to a string later.
                Always a string on un-raised error (the unparsed function string).

        Raises:
            ParsingError, any: A `ParsingError` if the function could not be
            found, otherwise error from function definition. Only raised if
            `raise_errors` is `True`

        Notes:
            The kwargs passed into the callable will be a mixture of the
            `default_kwargs` passed into `FuncParser.__init__`, kwargs given
            directly in the `$funcdef` string, and the `reserved_kwargs` this
            function gets from `.parse()`. For colliding keys, funcdef-defined
            kwargs will override default kwargs while reserved kwargs will always
            override the other two.

        """
        funcname, args, kwargs = parsedfunc.get()
        func = self.callables.get(funcname)

        if not func:
            if raise_errors:
                available = ", ".join(f"'{key}'" for key in self.callables)
                raise ParsingError(
                    f"Unknown parsed function '{str(parsedfunc)}' (available: {available})"
                )
            return str(parsedfunc)

        # build kwargs in the proper priority order
        kwargs = {
            **self.default_kwargs,
            **kwargs,
            **reserved_kwargs,
            **{"funcparser": self, "raise_errors": raise_errors},
        }

        try:
            ret = func(*args, **kwargs)
            return ret
        except ParsingError:
            if raise_errors:
                raise
            return str(parsedfunc)
        except Exception:
            logger.log_trace()
            if raise_errors:
                raise
            return str(parsedfunc)

    def parse(
        self,
        string,
        raise_errors=False,
        escape=False,
        strip=False,
        return_str=True,
        **reserved_kwargs,
    ):
        """
        Use parser to parse a string that may or may not have
        `$funcname(*args, **kwargs)` - style tokens in it. Only the callables
        used to initiate the parser will be eligible for parsing.

        Args:
            string (str): The string to parse.
            raise_errors (bool, optional): By default, a failing parse just
                means not parsing the string but leaving it as-is. If this is
                `True`, errors (like not closing brackets) will lead to an
                ParsingError.
            escape (bool, optional): If set, escape all found functions so they
                are not executed by later parsing.
            strip (bool, optional): If set, strip any inline funcs from string
                as if they were not there.
            return_str (bool, optional): If set (default), always convert the
                parse result to a string, otherwise return the result of the
                latest called inlinefunc (if called separately).
            **reserved_kwargs: If given, these are guaranteed to _always_ pass
                as part of each parsed callable's **kwargs. These  override
                same-named default options given in `__init__` as well as any
                same-named kwarg given in the string function. This is because
                it is often used by Evennia to pass necessary kwargs into each
                callable (like the current Session object for inlinefuncs).

        Returns:
            str or any: The parsed string, or the same string on error (if
                `raise_errors` is `False`). This is always a string

        Raises:
            ParsingError: If a problem is encountered and `raise_errors` is True.

        """
        start_char = self.start_char
        escape_char = self.escape_char

        # replace e.g. $$ with \$ so we only need to handle one escape method
        string = string.replace(start_char + start_char, escape_char + start_char)

        # parsing state
        callstack = []

        double_quoted = -1
        open_lparens = 0  # open (
        open_lsquare = 0  # open [
        open_lcurly = 0  # open {
        escaped = False
        current_kwarg = ""
        exec_return = ""

        curr_func = None
        fullstr = ""  # final string
        infuncstr = ""  # string parts inside the current level of $funcdef (including $)
        literal_infuncstr = False

        for char in string:
            if escaped:
                # always store escaped characters verbatim
                if curr_func:
                    infuncstr += char
                else:
                    fullstr += char
                escaped = False
                continue

            if char == escape_char:
                # don't store the escape-char itself
                escaped = True
                continue

            if char == start_char:
                # start a new function definition (not escaped as $$)

                if curr_func:
                    # we are starting a nested funcdef
                    if len(callstack) >= _MAX_NESTING - 1:
                        # stack full - ignore this function
                        if raise_errors:
                            raise ParsingError(
                                "Only allows for parsing nesting function defs "
                                f"to a max depth of {_MAX_NESTING}."
                            )
                        infuncstr += char
                        continue
                    else:
                        # store state for the current func and stack it
                        curr_func.current_kwarg = current_kwarg
                        curr_func.infuncstr = infuncstr
                        curr_func.double_quoted = double_quoted
                        curr_func.open_lparens = open_lparens
                        curr_func.open_lsquare = open_lsquare
                        curr_func.open_lcurly = open_lcurly
                        # we must strip the remaining funcstr so it's not counted twice
                        curr_func.rawstr = curr_func.rawstr[: -len(infuncstr)]
                        current_kwarg = ""
                        infuncstr = ""
                        double_quoted = -1
                        open_lparens = 0
                        open_lsquare = 0
                        open_lcurly = 0
                        exec_return = ""
                        literal_infuncstr = False
                        callstack.append(curr_func)

                # start a new func
                curr_func = _ParsedFunc(prefix=char, fullstr=char)
                continue

            if not curr_func:
                # a normal piece of string
                fullstr += char
                # this must always be a string
                return_str = True
                continue

            # in a function def (can be nested)

            curr_func.rawstr += char

            if exec_return != "" and char not in (",=)"):
                # if exec_return is followed by any other character
                # than one demarking an arg,kwarg or function-end
                # it must immediately merge as a string
                infuncstr += str(exec_return)
                exec_return = ""

            if char == '"':  # note that this is the same as '\"'
                # a double quote = flip status
                if double_quoted == 0:
                    infuncstr = infuncstr[1:]
                    double_quoted = -1
                elif double_quoted > 0:
                    prefix = infuncstr[0:double_quoted]
                    infuncstr = prefix + infuncstr[double_quoted + 1 :]
                    double_quoted = -1
                else:
                    infuncstr += char
                    infuncstr = infuncstr.strip()
                    double_quoted = len(infuncstr) - 1
                    literal_infuncstr = True

                continue

            if double_quoted >= 0:
                # inside a string definition - this escapes everything else
                infuncstr += char
                continue

            # special characters detected inside function def
            if char == "(":
                if not curr_func.funcname:
                    # end of a funcdef name
                    curr_func.funcname = infuncstr
                    curr_func.fullstr += infuncstr + char
                    infuncstr = ""
                else:
                    # just a random left-parenthesis
                    infuncstr += char
                # track the open left-parenthesis
                open_lparens += 1
                continue

            if char in "[]":
                # a square bracket - start/end of a list?
                infuncstr += char
                open_lsquare += -1 if char == "]" else 1
                continue

            if char in "{}":
                # a curly bracket - start/end of dict/set?
                infuncstr += char
                open_lcurly += -1 if char == "}" else 1
                continue

            if char == "=":
                # beginning of a keyword argument
                if exec_return != "":
                    infuncstr = exec_return
                current_kwarg = infuncstr.strip()
                curr_func.kwargs[current_kwarg] = ""
                curr_func.fullstr += infuncstr + char
                infuncstr = ""
                continue

            if char in (",)"):
                # commas and right-parens may indicate arguments ending

                if open_lparens > 1:
                    # one open left-parens is ok (beginning of arglist), more
                    # indicate we are inside an unclosed, nested (, so
                    # we need to not count this as a new arg or end of funcdef.
                    infuncstr += char
                    open_lparens -= 1 if char == ")" else 0
                    continue

                if open_lcurly > 0 or open_lsquare > 0:
                    # also escape inside an open [... or {... structure
                    infuncstr += char
                    continue

                if exec_return != "":
                    # store the execution return as-received
                    if current_kwarg:
                        curr_func.kwargs[current_kwarg] = exec_return
                    else:
                        curr_func.args.append(exec_return)
                else:
                    if not literal_infuncstr:
                        infuncstr = infuncstr.strip()

                    # store a string instead
                    if current_kwarg:
                        curr_func.kwargs[current_kwarg] = infuncstr
                    elif literal_infuncstr or infuncstr.strip():
                        # don't store the empty string
                        curr_func.args.append(infuncstr)

                # note that at this point either exec_return or infuncstr will
                # be empty. We need to store the full string so we can print
                # it 'raw' in case this funcdef turns out to e.g. lack an
                # ending paranthesis
                curr_func.fullstr += str(exec_return) + infuncstr + char

                current_kwarg = ""
                exec_return = ""
                infuncstr = ""
                literal_infuncstr = False

                if char == ")":
                    # closing the function list - this means we have a
                    # ready function-def to run.
                    open_lparens = 0

                    if strip:
                        # remove function as if it returned empty
                        exec_return = ""
                    elif escape:
                        # get function and set it as escaped
                        exec_return = escape_char + curr_func.fullstr
                    else:
                        # execute the function - the result may be a string or
                        # something else
                        exec_return = self.execute(
                            curr_func, raise_errors=raise_errors, **reserved_kwargs
                        )

                    if callstack:
                        # unnest the higher-level funcdef from stack
                        # and continue where we were
                        curr_func = callstack.pop()
                        current_kwarg = curr_func.current_kwarg
                        if curr_func.infuncstr:
                            # if we have an ongoing string, we must merge the
                            # exec into this as a part of that string
                            infuncstr = curr_func.infuncstr + str(exec_return)
                            exec_return = ""
                        curr_func.infuncstr = ""
                        double_quoted = curr_func.double_quoted
                        open_lparens = curr_func.open_lparens
                        open_lsquare = curr_func.open_lsquare
                        open_lcurly = curr_func.open_lcurly
                    else:
                        # back to the top-level string - this means the
                        # exec_return should always be converted to a string.
                        curr_func = None
                        fullstr += str(exec_return)
                        if return_str:
                            exec_return = ""
                        infuncstr = ""
                        literal_infuncstr = False
                continue

            infuncstr += char

        if curr_func:
            # if there is a still open funcdef or defs remaining in callstack,
            # these are malformed (no closing bracket) and we should get their
            # strings as-is.
            callstack.append(curr_func)
            for inum, _ in enumerate(range(len(callstack))):
                funcstr = str(callstack.pop())
                if inum == 0 and funcstr.endswith(infuncstr):
                    # avoid double-echo of nested function calls. This should
                    # produce a good result most of the time, but it's not 100%
                    # guaranteed to, since it can ignore genuine duplicates
                    infuncstr = funcstr
                else:
                    infuncstr = funcstr + infuncstr

        if not return_str and exec_return != "":
            # return explicit return
            return exec_return

        # add the last bit to the finished string
        fullstr += infuncstr

        return fullstr

    def parse_to_any(
        self, string, raise_errors=False, escape=False, strip=False, **reserved_kwargs
    ):
        """
        This parses a string and if the string only contains a "$func(...)",
        the return will be the return value of that function, even if it's not
        a string. If mixed in with other strings, the result will still always
        be a string.

        Args:
            string (str): The string to parse.
            raise_errors (bool, optional): If unset, leave a failing (or
                unrecognized) inline function as unparsed in the string. If set,
                raise an ParsingError.
            escape (bool, optional): If set, escape all found functions so they
                are not executed by later parsing.
            strip (bool, optional): If set, strip any inline funcs from string
                as if they were not there.
            **reserved_kwargs: If given, these are guaranteed to _always_ pass
                as part of each parsed callable's **kwargs. These  override
                same-named default options given in `__init__` as well as any
                same-named kwarg given in the string function. This is because
                it is often used by Evennia to pass necessary kwargs into each
                callable (like the current Session object for inlinefuncs).

        Returns:
            any: The return from the callable. Or string if the callable is not
                given alone in the string.

        Raises:
            ParsingError: If a problem is encountered and `raise_errors` is True.

        Notes:
            This is a convenience wrapper for `self.parse(..., return_str=False)` which
            accomplishes the same thing.

        Examples:
            ::

                from ast import literal_eval
                from evennia.utils.funcparser import FuncParser


                def ret1(*args, **kwargs):
                    return 1

                parser = FuncParser({"lit": lit})

                assert parser.parse_to_any("$ret1()" == 1
                assert parser.parse_to_any("$ret1() and text" == '1 and text'

        """
        return self.parse(
            string,
            raise_errors=raise_errors,
            escape=escape,
            strip=strip,
            return_str=False,
            **reserved_kwargs,
        )


#
# Default funcparser callables. These are made available from this module's
# FUNCPARSER_CALLABLES.
#


def funcparser_callable_eval(*args, **kwargs):
    """
    Funcparser callable. This will combine safe evaluations to try to parse the
    incoming string into a python object. If it fails, the return will be same
    as the input.

    Args:
        string (str): The string to parse. Only simple literals or operators are allowed.

    Returns:
        any: The string parsed into its Python form, or the same as input.

    Examples:
        - `$py(1) -> 1`
        - `$py([1,2,3,4] -> [1, 2, 3]`
        - `$py(3 + 4) -> 7`

    """
    args, kwargs = safe_convert_to_types(("py", {}), *args, **kwargs)
    return args[0] if args else ""


def funcparser_callable_toint(*args, **kwargs):
    """Usage: $toint(43.0) -> 43"""
    inp = funcparser_callable_eval(*args, **kwargs)
    try:
        return int(inp)
    except TypeError:
        return inp
    except ValueError:
        return inp


def funcparser_callable_int2str(*args, **kwargs):
    """
    Usage: $int2str(1) -> 'one' etc, up to 12->twelve.

    Args:
        number (int): The number. If not an int, will be converted.

    Uses the int2str utility function.
    """
    if not args:
        return ""
    try:
        number = int(args[0])
    except ValueError:
        return args[0]
    return int2str(number)


def funcparser_callable_an(*args, **kwargs):
    """
    Usage: $an(thing) -> a thing

    Adds a/an depending on if the first letter of the given word is a consonant or not.

    """
    if not args:
        return ""
    item = str(args[0])
    if item and item[0] in "aeiouyAEIOUY":
        return f"an {item}"
    return f"a {item}"


def _apply_operation_two_elements(*args, operator="+", **kwargs):
    """
    Helper operating on two arguments

    Args:
        val1 (any): First value to operate on.
        val2 (any): Second value to operate on.

    Return:
        any: The result of val1 + val2. Values must be
            valid simple Python structures possible to add,
            such as numbers, lists etc. The $eval is usually
            better for non-list arithmetic.

    """
    args, kwargs = safe_convert_to_types((("py", "py"), {}), *args, **kwargs)
    if not len(args) > 1:
        return ""
    val1, val2 = args[0], args[1]
    try:
        if operator == "+":
            return val1 + val2
        elif operator == "-":
            return val1 - val2
        elif operator == "*":
            return val1 * val2
        elif operator == "/":
            return val1 / val2
    except Exception:
        if kwargs.get("raise_errors"):
            raise
        return ""


def funcparser_callable_add(*args, **kwargs):
    """Usage: `$add(val1, val2) -> val1 + val2`"""
    return _apply_operation_two_elements(*args, operator="+", **kwargs)


def funcparser_callable_sub(*args, **kwargs):
    """Usage: ``$sub(val1, val2) -> val1 - val2`"""
    return _apply_operation_two_elements(*args, operator="-", **kwargs)


def funcparser_callable_mult(*args, **kwargs):
    """Usage: `$mult(val1, val2) -> val1 * val2`"""
    return _apply_operation_two_elements(*args, operator="*", **kwargs)


def funcparser_callable_div(*args, **kwargs):
    """Usage: `$mult(val1, val2) -> val1 / val2`"""
    return _apply_operation_two_elements(*args, operator="/", **kwargs)


def funcparser_callable_round(*args, **kwargs):
    """
    Funcparser callable. Rounds an incoming float to a
    certain number of significant digits.

    Args:
        inp (str or number): If a string, it will attempt
            to be converted to a number first.
        significant (int): The number of significant digits.  Default is None -
            this will turn the result into an int.

    Returns:
        any: The rounded value or inp if inp was not a number.

    Examples:
        - `$round(3.5434343, 3) -> 3.543`
        - `$round($random(), 2)` - rounds random result, e.g `0.22`

    """
    if not args:
        return ""
    args, _ = safe_convert_to_types(((float, int), {}), *args, **kwargs)

    num, *significant = args
    significant = significant[0] if significant else 0
    try:
        return round(num, significant)
    except Exception:
        if kwargs.get("raise_errors"):
            raise
        return ""


def funcparser_callable_random(*args, **kwargs):
    """
    Funcparser callable. Returns a random number between 0 and 1, from 0 to a
    maximum value, or within a given range (inclusive).

    Args:
        minval (str, optional): Minimum value. If not given, assumed 0.
        maxval (str, optional): Maximum value.

    Notes:
        If either of the min/maxvalue has a '.' in it, a floating-point random
        value will be returned. Otherwise it will be an
        integer value in the given range.

    Examples:
        - `$random()` - random value [0 .. 1) (float).
        - `$random(5)` - random value [0..5] (int)
        - `$random(5.0)` - random value [0..5] (float)
        - `$random(5, 10)` - random value [5..10] (int)
        - `$random(5, 10.0)` - random value [5..10] (float)

    """
    args, _ = safe_convert_to_types((("py", "py"), {}), *args, **kwargs)

    nargs = len(args)
    if nargs == 1:
        # only maxval given
        minval, maxval = 0, args[0]
    elif nargs > 1:
        minval, maxval = args[:2]
    else:
        minval, maxval = 0, 1

    try:
        if isinstance(minval, float) or isinstance(maxval, float):
            return minval + ((maxval - minval) * random.random())
        else:
            return random.randint(minval, maxval)
    except Exception:
        if kwargs.get("raise_errors"):
            raise
        return ""


def funcparser_callable_randint(*args, **kwargs):
    """
    Usage: $randint(start, end):

    Legacy alias - always returns integers.

    """
    return int(funcparser_callable_random(*args, **kwargs))


def funcparser_callable_choice(*args, **kwargs):
    """
    FuncParser callable. Picks a random choice from a list.

    Args:
        listing (list): A list of items to randomly choose between.
            This will be converted from a string to a real list.
        *args: If multiple args are given, will pick one randomly from them.

    Returns:
        any: The randomly chosen element.

    Example:
        - `$choice(key, flower, house)`
        - `$choice([1, 2, 3, 4])`

    """
    if not args:
        return ""

    nargs = len(args)
    if nargs == 1:
        # this needs to be a list/tuple for this to make sense
        args, _ = safe_convert_to_types(("py", {}), args[0], **kwargs)
        args = make_iter(args[0]) if args else None
    else:
        # separate arg per entry
        converters = ["py" for _ in range(nargs)]
        args, _ = safe_convert_to_types((converters, {}), *args, **kwargs)

    if not args:
        return ""
    try:
        return random.choice(args)
    except Exception:
        if kwargs.get("raise_errors"):
            raise
        return ""


def funcparser_callable_pad(*args, **kwargs):
    """
    FuncParser callable. Pads text to given width, optionally with fill-characters

    Args:
        text (str): Text to pad.
        width (int): Width of padding.
        align (str, optional): Alignment of padding; one of 'c', 'l' or 'r'.
        fillchar (str, optional): Character used for padding. Defaults to a space.

    Example:
        - `$pad(text, 12, r, ' ') -> "        text"`
        - `$pad(text, width=12, align=c, fillchar=-) -> "----text----"`

    """
    if not args:
        return ""
    args, kwargs = safe_convert_to_types(
        ((str, int, str, str), {"width": int, "align": str, "fillchar": str}), *args, **kwargs
    )

    text, *rest = args
    nrest = len(rest)
    try:
        width = int(kwargs.get("width", rest[0] if nrest > 0 else _CLIENT_DEFAULT_WIDTH))
    except (TypeError, ValueError):
        width = _CLIENT_DEFAULT_WIDTH

    align = kwargs.get("align", rest[1] if nrest > 1 else "c")
    fillchar = kwargs.get("fillchar", rest[2] if nrest > 2 else " ")
    if align not in ("c", "l", "r"):
        align = "c"
    return pad(str(text), width=width, align=align, fillchar=fillchar)


def funcparser_callable_crop(*args, **kwargs):
    """
    FuncParser callable. Crops ingoing text to given widths.

    Args:
        text (str, optional): Text to crop.
        width (str, optional): Will be converted to an integer. Width of
            crop in characters.
        suffix (str, optional): End string to mark the fact that a part
            of the string was cropped. Defaults to `[...]`.

    Example:
        - `$crop(A long text, 10, [...]) -> "A lon[...]"`
        - `$crop(text, width=11, suffix='[...]) -> "A long[...]"`

    """
    if not args:
        return ""
    text, *rest = args
    nrest = len(rest)
    try:
        width = int(kwargs.get("width", rest[0] if nrest > 0 else _CLIENT_DEFAULT_WIDTH))
    except (TypeError, ValueError):
        width = _CLIENT_DEFAULT_WIDTH
    suffix = kwargs.get("suffix", rest[1] if nrest > 1 else "[...]")
    return crop(str(text), width=width, suffix=str(suffix))


def funcparser_callable_space(*args, **kwarg):
    """
    Usage: $space(43)

    Insert a length of space.

    """
    if not args:
        return ""
    try:
        width = int(args[0])
    except ValueError:
        width = 1
    return " " * width


def funcparser_callable_justify(*args, **kwargs):
    """
    Justify text across a width, default across screen width.

    Args:
        text (str): Text to justify.
        width (int, optional): Defaults to default screen width.
        align (str, optional): One of 'l', 'c', 'r' or 'f' for 'full'.
        indent (int, optional): Intendation of text block, if any.

    Returns:
        str: The justified text.

    Examples:
        - `$just(text, width=40)`
        - `$just(text, align=r, indent=2)`

    """
    if not args:
        return ""
    text, *rest = args
    lrest = len(rest)
    try:
        width = int(kwargs.get("width", rest[0] if lrest > 0 else _CLIENT_DEFAULT_WIDTH))
    except (TypeError, ValueError):
        width = _CLIENT_DEFAULT_WIDTH
    align = str(kwargs.get("align", rest[1] if lrest > 1 else "f"))
    try:
        indent = int(kwargs.get("indent", rest[2] if lrest > 2 else 0))
    except (TypeError, ValueError):
        indent = 0
    return justify(str(text), width=width, align=align, indent=indent)


# legacy for backwards compatibility
def funcparser_callable_left_justify(*args, **kwargs):
    "Usage: $ljust(text)"
    return funcparser_callable_justify(*args, align="l", **kwargs)


def funcparser_callable_right_justify(*args, **kwargs):
    "Usage: $rjust(text)"
    return funcparser_callable_justify(*args, align="r", **kwargs)


def funcparser_callable_center_justify(*args, **kwargs):
    "Usage: $cjust(text)"
    return funcparser_callable_justify(*args, align="c", **kwargs)


def funcparser_callable_clr(*args, **kwargs):
    """
    FuncParser callable. Colorizes nested text.

    Args:
        startclr (str, optional): An ANSI color abbreviation without the
            prefix `|`, such as `r` (red foreground) or `[r` (red background).
        text (str, optional): Text
        endclr (str, optional): The color to use at the end of the string. Defaults
            to `|n` (reset-color).
    Kwargs:
        color (str, optional): If given,

    Example:
        - `$clr(r, text, n) -> "|rtext|n"`
        - `$clr(r, text) -> "|rtext|n`
        - `$clr(text, start=r, end=n) -> "|rtext|n"`

    """
    if not args:
        return ""

    startclr, text, endclr = "", "", ""
    if len(args) > 1:
        # $clr(pre, text, post))
        startclr, *rest = args
        if rest:
            text, *endclr = rest
            if endclr:
                endclr = endclr[0]
    else:
        # $clr(text, start=pre, end=post)
        text = args[0]
        startclr = kwargs.get("start", "")
        endclr = kwargs.get("end", "")

    startclr = "|" + startclr if startclr else ""
    endclr = "|" + endclr if endclr else ("|n" if startclr else "")
    return f"{startclr}{text}{endclr}"


def funcparser_callable_pluralize(*args, **kwargs):
    """
    FuncParser callable. Handles pluralization of a word.

    Args:
        singular_word (str): The base (singular) word to optionally pluralize
        number (int): The number of elements; if 1 (or 0), use `singular_word` as-is,
            otherwise use plural form.
        plural_word (str, optional): If given, this will be used if `number`
            is greater than one. If not given, we simply add 's' to the end of
            `singular_word`.

    Example:
        - `$pluralize(thing, 2)` -> "things"
        - `$pluralize(goose, 18, geese)` -> "geese"

    """
    if not args:
        return ""
    nargs = len(args)
    if nargs > 2:
        singular_word, number, plural_word = args[:3]
    elif nargs > 1:
        singular_word, number = args[:2]
        plural_word = f"{singular_word}s"
    else:
        singular_word, number = args[0], 1
    return singular_word if abs(int(number)) in (0, 1) else plural_word


def funcparser_callable_search(*args, caller=None, access="control", **kwargs):
    """
    FuncParser callable. Finds an object based on name or #dbref. Note that
    this requries the parser be called with the caller's Session for proper
    security. If called without session, the call is aborted.

    Args:
        query (str): The key or dbref to search for. This can consist of any args used
            for one of the regular search methods. Also kwargs will be passed into
            the search (except the kwargs given below)

    Keyword Args:
        return_list (bool): If set, return a list of objects with
            0, 1 or more matches to `query`. Defaults to False.
        type (str): One of 'obj', 'account', 'script'
        caller (Entity): Supplied to Parser. This is required and will
            be passed into the access check for the entity being searched for.
            The 'control' permission is required.
        access (str): Which locktype access to check. Unset to disable the
            security check.
        **kwargs: Will be passed into the main search.

    Returns:
        any: An entity match or None if no match or a list if `return_list` is set.

    Raise:
        ParsingError: If zero/multimatch and `return_list` is False, or caller was not
            passed into parser.

    Examples:
        - "$search(#233)"
        - "$search(Tom, type=account)"
        - "$search(meadow, return_list=True)"
        - "$search(beach, category=outdoors, type=tag)

    """
    # clean out funcparser-specific kwargs so we can use the kwargs for
    # searching
    search_kwargs = {
        key: value
        for key, value in kwargs.items()
        if key not in ("funcparser", "raise_errors", "type", "return_list")
    }
    return_list = str(kwargs.pop("return_list", "false")).lower() == "true"

    if not args:
        return [] if return_list else None
    if not caller:
        raise ParsingError("$search requires a `caller` passed to the parser.")

    typ = kwargs.get("type", "obj")
    targets = []
    if typ == "obj":
        targets = search.search_object(*args, **search_kwargs)
    elif typ == "account":
        targets = search.search_account(*args, **search_kwargs)
    elif typ == "script":
        targets = search.search_script(*args, **search_kwargs)
    elif typ == "tag":
        targets = search.search_object_by_tag(*args, **search_kwargs)

    if not targets:
        if return_list:
            return []
        raise ParsingError(f"$search: Query '{args[0]}' gave no matches.")

    if len(targets) > 1 and not return_list:
        raise ParsingError(
            "$search: Query '{query}' found {num} matches. "
            "Set return_list=True to accept a list".format(query=query, num=len(targets))
        )

    for target in targets:
        if not target.access(caller, access):
            raise ParsingError("$search Cannot add found entity - access failure.")

    return list(targets) if return_list else targets[0]


def funcparser_callable_search_list(*args, caller=None, access="control", **kwargs):
    """
    Usage: $objlist(#123)

    Legacy alias for search with a return_list=True kwarg preset.

    """
    return funcparser_callable_search(
        *args, caller=caller, access=access, return_list=True, **kwargs
    )


def funcparser_callable_you(
    *args, caller=None, receiver=None, mapping=None, capitalize=False, **kwargs
):
    """
    Usage: $you() or $you(key)

    Replaces with you for the caller of the string, with the display_name
    of the caller for others.

    Keyword Args:
        caller (Object): The 'you' in the string. This is used unless another
            you-key is passed to the callable in combination with `mapping`.
        receiver (Object): The recipient of the string.
        mapping (dict, optional): This is a mapping `{key:Object, ...}` and is
            used to find which object `$you(key)` refers to. If not given, the
            `caller` kwarg is used.
        capitalize (bool): Passed by the You helper, to capitalize you.

    Returns:
        str: The parsed string.

    Raises:
        ParsingError: If `caller` and `receiver` were not supplied.

    Notes:
        The kwargs should be passed the to parser directly.

    Examples:
        This can be used by the say or emote hooks to pass actor stance
        strings. This should usually be combined with the $conj() callable.

        - `With a grin, $you() $conj(jump) at $you(tommy).`

        The caller-object will see "With a grin, you jump at Tommy."
        Tommy will see "With a grin, CharName jumps at you."
        Others will see "With a grin, CharName jumps at Tommy."

    """
    if args and mapping:
        # this would mean a $you(key) form
        caller = mapping.get(args[0], None)

    if not (caller and receiver):
        raise ParsingError("No caller or receiver supplied to $you callable.")

    capitalize = bool(capitalize)
    if caller == receiver:
        return "You" if capitalize else "you"
    return (
        caller.get_display_name(looker=receiver)
        if hasattr(caller, "get_display_name")
        else str(caller)
    )


def funcparser_callable_you_capitalize(
    *args, you=None, receiver=None, mapping=None, capitalize=True, **kwargs
):
    """
    Usage: $You() - capitalizes the 'you' output.

    """
    return funcparser_callable_you(
        *args, you=you, receiver=receiver, mapping=mapping, capitalize=capitalize, **kwargs
    )


def funcparser_callable_your(
    *args, caller=None, receiver=None, mapping=None, capitalize=False, **kwargs
):
    """
    Usage: $your() or $your(key)

    Replaces with your for the caller of the string, with the display_name +'s
    of the caller for others.

    Keyword Args:
        caller (Object): The 'your' in the string. This is used unless another
            your-key is passed to the callable in combination with `mapping`.
        receiver (Object): The recipient of the string.
        mapping (dict, optional): This is a mapping `{key:Object, ...}` and is
            used to find which object `$you(key)` refers to. If not given, the
            `caller` kwarg is used.
        capitalize (bool): Passed by the You helper, to capitalize you.

    Returns:
        str: The parsed string.

    Raises:
        ParsingError: If `caller` and `receiver` were not supplied.

    Notes:
        The kwargs should be passed the to parser directly.

    Examples:
        This can be used by the say or emote hooks to pass actor stance
        strings.

        - `$your() pet jumps at $you(tommy).`

        The caller-object will see "Your pet jumps Tommy."
        Tommy will see "CharName's pet jumps at you."
        Others will see "CharName's pet jumps at Tommy."

    """
    if args and mapping:
        # this would mean a $your(key) form
        caller = mapping.get(args[0], None)

    if not (caller and receiver):
        raise ParsingError("No caller or receiver supplied to $your callable.")

    capitalize = bool(capitalize)
    if caller == receiver:
        return "Your" if capitalize else "your"

    name = (
        caller.get_display_name(looker=receiver)
        if hasattr(caller, "get_display_name")
        else str(caller)
    )

    return name + "'s"


def funcparser_callable_your_capitalize(
    *args, you=None, receiver=None, mapping=None, capitalize=True, **kwargs
):
    """
    Usage: $Your() - capitalizes the 'your' output.

    """
    return funcparser_callable_your(
        *args, you=you, receiver=receiver, mapping=mapping, capitalize=capitalize, **kwargs
    )


def funcparser_callable_conjugate(*args, caller=None, receiver=None, **kwargs):
    """
    Usage: $conj(word, [options])

    Conjugate a verb according to if it should be 2nd or third person.

    Keyword Args:
        caller (Object): The object who represents 'you' in the string.
        receiver (Object): The recipient of the string.

    Returns:
        str: The parsed string.

    Raises:
        ParsingError: If `you` and `recipient` were not both supplied.

    Notes:
        Note that the verb will not be capitalized. It also
        assumes that the active party (You) is the one performing the verb.
        This automatic conjugation will fail if the active part is another person
        than 'you'. The caller/receiver must be passed to the parser directly.

    Examples:
        This is often used in combination with the $you/You( callables.

        - `With a grin, $you() $conj(jump)`

        You will see "With a grin, you jump."
        Others will see "With a grin, CharName jumps."

    """
    if not args:
        return ""
    if not (caller and receiver):
        raise ParsingError("No caller/receiver supplied to $conj callable")

    second_person_str, third_person_str = verb_actor_stance_components(args[0])
    return second_person_str if caller == receiver else third_person_str


def funcparser_callable_pronoun(*args, caller=None, receiver=None, capitalize=False, **kwargs):
    """

    Usage: $pron(word, [options])

    Adjust pronouns to the expected form. Pronouns are words you use instead of a
    proper name, such as 'him', 'herself', 'theirs' etc. These look different
    depending on who sees the outgoing string.

    The parser maps between this table ...

    ====================  =======  =======  ==========  ==========  ===========
    1st/2nd person        Subject  Object   Possessive  Possessive  Reflexive
                          Pronoun  Pronoun  Adjective   Pronoun     Pronoun
    ====================  =======  =======  ==========  ==========  ===========
    1st person               I        me        my        mine       myself
    1st person plural       we       us        our        ours       ourselves
    2nd person              you      you       your       yours      yourself
    2nd person plural       you      you       your       yours      yourselves
    ====================  =======  =======  ==========  ==========  ===========

    ... and this table (and vice versa).

    ====================  =======  =======  ==========  ==========  ===========
    3rd person            Subject  Object   Possessive  Possessive  Reflexive
                          Pronoun  Pronoun  Adjective   Pronoun     Pronoun
    ====================  =======  =======  ==========  ==========  ===========
    3rd person male         he       him       his        his        himself
    3rd person female       she      her       her        hers       herself
    3rd person neutral      it       it        its                   itself
    3rd person plural       they    them       their      theirs     themselves
    ====================  =======  =======  ==========  ==========  ===========

    This system will examine `caller` for either a property or a callable `.gender` to
    get a default gender fallback (if not specified in the call). If a callable,
    `.gender` will be called without arguments and should return a string
    `male`/`female`/`neutral`/`plural` (plural is considered a gender for this purpose).
    If no `gender` property/callable is found, `neutral` is used as a fallback.

    The pronoun-type default (if not specified in call) is `subject pronoun`.

    Args:
        pronoun (str): Input argument to parsed call. This can be any of the pronouns
            in the table above. If given in 1st/second form, they will be mappped to
            3rd-person form for others viewing the message (but will need extra input
            via the `gender`, see below). If given on 3rd person form, this will be
            mapped to 2nd person form for `caller` unless `viewpoint` is specified
            in options.
        options (str, optional): A space- or comma-separated string detailing `pronoun_type`,
            `gender`/`plural` and/or `viewpoint` to help the mapper differentiate between
            non-unique cases (such as if `you` should become `him` or `they`).
            Allowed values are:

            - `subject pronoun`/`subject`/`sp` (I, you, he, they)
            - `object pronoun`/`object/`/`op`  (me, you, him, them)
            - `possessive adjective`/`adjective`/`pa` (my, your, his, their)
            - `possessive pronoun`/`pronoun`/`pp`  (mine, yours, his, theirs)
            - `male`/`m`
            - `female`/`f`
            - `neutral`/`n`
            - `plural`/`p`
            - `1st person`/`1st`/`1`
            - `2nd person`/`2nd`/`2`
            - `3rd person`/`3rd`/`3`

    Keyword Args:

        caller (Object): The object creating the string. If this has a property 'gender',
            it will be checked for a string 'male/female/neutral' to determine
            the 3rd person gender (but if `pronoun_type` contains a gender
            component, that takes precedence). Provided automatically to the
            funcparser.
        receiver (Object): The recipient of the string. This being the same as
            `caller` or not helps determine 2nd vs 3rd-person forms. This is
            provided automatically by the funcparser.
        capitalize (bool): The input retains its capitalization. If this is set the output is
            always capitalized.

    Examples:

        ======================  =============    ===========
        Input                   caller sees      others see
        ======================  =============    ===========
        $pron(I, m)             I                he
        $pron(you,fo)           you              her
        $pron(yourself)         yourself         itself
        $pron(its)              your             its
        $pron(you,op,p)         you              them
        ======================  =============    ===========

    Notes:
        There is no option to specify reflexive pronouns since they are all unique
        and the mapping can always be auto-detected.

    """
    if not args:
        return ""

    pronoun, *options = args
    # options is either multiple args or a space-separated string
    if len(options) == 1:
        options = options[0]

    # default differentiators
    default_pronoun_type = "subject pronoun"
    default_gender = "neutral"
    default_viewpoint = "2nd person"

    if hasattr(caller, "gender"):
        if callable(caller.gender):
            default_gender = caller.gender()
        else:
            default_gender = caller.gender

    if "viewpoint" in kwargs:
        # passed into FuncParser initialization
        default_viewpoint = kwargs["viewpoint"]

    pronoun_1st_or_2nd_person, pronoun_3rd_person = pronoun_to_viewpoints(
        pronoun,
        options,
        pronoun_type=default_pronoun_type,
        gender=default_gender,
        viewpoint=default_viewpoint,
    )

    if capitalize:
        pronoun_1st_or_2nd_person = pronoun_1st_or_2nd_person.capitalize()
        pronoun_3rd_person = pronoun_3rd_person.capitalize()

    return pronoun_1st_or_2nd_person if caller == receiver else pronoun_3rd_person


def funcparser_callable_pronoun_capitalize(
    *args, caller=None, receiver=None, capitalize=True, **kwargs
):
    """
    Usage: $Pron(word, [options]) - always maps to a capitalized word.

    """
    return funcparser_callable_pronoun(
        *args, caller=caller, receiver=receiver, capitalize=capitalize, **kwargs
    )


# these are made available as callables by adding 'evennia.utils.funcparser' as
# a callable-path when initializing the FuncParser.

FUNCPARSER_CALLABLES = {
    # 'standard' callables
    # eval and arithmetic
    "eval": funcparser_callable_eval,
    "add": funcparser_callable_add,
    "sub": funcparser_callable_sub,
    "mult": funcparser_callable_mult,
    "div": funcparser_callable_div,
    "round": funcparser_callable_round,
    "toint": funcparser_callable_toint,
    # randomizers
    "random": funcparser_callable_random,
    "randint": funcparser_callable_randint,
    "choice": funcparser_callable_choice,
    # string manip
    "pad": funcparser_callable_pad,
    "crop": funcparser_callable_crop,
    "just": funcparser_callable_justify,
    "ljust": funcparser_callable_left_justify,
    "rjust": funcparser_callable_right_justify,
    "cjust": funcparser_callable_center_justify,
    "justify": funcparser_callable_justify,  # aliases for backwards compat
    "justify_left": funcparser_callable_left_justify,
    "justify_right": funcparser_callable_right_justify,
    "justify_center": funcparser_callable_center_justify,
    "space": funcparser_callable_space,
    "clr": funcparser_callable_clr,
    "pluralize": funcparser_callable_pluralize,
    "int2str": funcparser_callable_int2str,
    "an": funcparser_callable_an,
}

SEARCHING_CALLABLES = {
    # requires `caller` and optionally `access` to be passed into parser
    "search": funcparser_callable_search,
    "obj": funcparser_callable_search,  # aliases for backwards compat
    "objlist": funcparser_callable_search_list,
    "dbref": funcparser_callable_search,
}

ACTOR_STANCE_CALLABLES = {
    # requires `you`, `receiver` and `mapping` to be passed into parser
    "you": funcparser_callable_you,
    "You": funcparser_callable_you_capitalize,
    "your": funcparser_callable_your,
    "Your": funcparser_callable_your_capitalize,
    "obj": funcparser_callable_you,
    "Obj": funcparser_callable_you_capitalize,
    "conj": funcparser_callable_conjugate,
    "pron": funcparser_callable_pronoun,
    "Pron": funcparser_callable_pronoun_capitalize,
    **FUNCPARSER_CALLABLES,
}



"""
File-based help entries. These complements command-based help and help entries
added in the database using the `sethelp` command in-game.

Control where Evennia reads these entries with `settings.FILE_HELP_ENTRY_MODULES`,
which is a list of python-paths to modules to read.

A module like this should hold a global `HELP_ENTRY_DICTS` list, containing
dicts that each represent a help entry. If no `HELP_ENTRY_DICTS` variable is
given, all top-level variables that are dicts in the module are read as help
entries.

Each dict is on the form
::

    {'key': <str>,
     'text': <str>}``     # the actual help text. Can contain # subtopic sections
     'category': <str>,   # optional, otherwise settings.DEFAULT_HELP_CATEGORY
     'aliases': <list>,   # optional
     'locks': <str>       # optional, 'view' controls seeing in help index, 'read'
                          #           if the entry can be read. If 'view' is unset,
                          #           'read' is used for the index. If unset, everyone
                          #           can read/view the entry.

"""

HELP_ENTRY_DICTS = [
    {
        "key": "evennia",
        "aliases": ["ev"],
        "category": "General",
        "locks": "read:perm(Developer)",
        "text": """
            Evennia is a MU-game server and framework written in Python. You can read more
            on https://www.evennia.com.

            # subtopics

            ## Installation

            You'll find installation instructions on https://www.evennia.com.

            ## Community

            There are many ways to get help and communicate with other devs!

            ### Discussions

            The Discussions forum is found at https://github.com/evennia/evennia/discussions.

            ### Discord

            There is also a discord channel for chatting - connect using the
            following link: https://discord.gg/AJJpcRUhtF

        """,
    },
]



"""
Color markups

Contribution, Griatch 2017

Additional color markup styles for Evennia (extending or replacing the default |r, |234 etc).


Installation:

Import the desired style variables from this module into mygame/server/conf/settings.py and add them
to these settings variables. Each are specified as a list, and multiple such lists can be added to
each variable to support multiple formats. Note that list order affects which regexes are applied
first. You must restart both Portal and Server for color tags to update.

Assign to the following settings variables:

    COLOR_ANSI_EXTRA_MAP - a mapping between regexes and ANSI colors
    COLOR_XTERM256_EXTRA_FG - regex for defining XTERM256 foreground colors
    COLOR_XTERM256_EXTRA_BG - regex for defining XTERM256 background colors
    COLOR_XTERM256_EXTRA_GFG - regex for defining XTERM256 grayscale foreground colors
    COLOR_XTERM256_EXTRA_GBG - regex for defining XTERM256 grayscale background colors
    COLOR_ANSI_BRIGHT_BG_EXTRA_MAP = ANSI does not support bright backgrounds; we fake
        this by mapping ANSI markup to matching bright XTERM256 backgrounds

    COLOR_NO_DEFAULT - Set True/False. If False (default), extend the default markup, otherwise
        replace it completely.


To add the {- "curly-bracket" style, add the following to your settings file, then reboot both
Server and Portal:

from evennia.contrib.base_systems import color_markups
COLOR_ANSI_EXTRA_MAP = color_markups.CURLY_COLOR_ANSI_EXTRA_MAP
COLOR_XTERM256_EXTRA_FG = color_markups.CURLY_COLOR_XTERM256_EXTRA_FG
COLOR_XTERM256_EXTRA_BG = color_markups.CURLY_COLOR_XTERM256_EXTRA_BG
COLOR_XTERM256_EXTRA_GFG = color_markups.CURLY_COLOR_XTERM256_EXTRA_GFG
COLOR_XTERM256_EXTRA_GBG = color_markups.CURLY_COLOR_XTERM256_EXTRA_GBG
COLOR_ANSI_BRIGHT_BG_EXTRA_MAP = color_markups.CURLY_COLOR_ANSI_BRIGHT_BG_EXTRA_MAP


To add the %c- "mux/mush" style, add the following to your settings file, then reboot both Server
and Portal:

from evennia.contrib.base_systems import color_markups
COLOR_ANSI_EXTRA_MAP = color_markups.MUX_COLOR_ANSI_EXTRA_MAP
COLOR_XTERM256_EXTRA_FG = color_markups.MUX_COLOR_XTERM256_EXTRA_FG
COLOR_XTERM256_EXTRA_BG = color_markups.MUX_COLOR_XTERM256_EXTRA_BG
COLOR_XTERM256_EXTRA_GFG = color_markups.MUX_COLOR_XTERM256_EXTRA_GFG
COLOR_XTERM256_EXTRA_GBG = color_markups.MUX_COLOR_XTERM256_EXTRA_GBG
COLOR_ANSI_BRIGHT_BGS_EXTRA_MAP = color_markups.CURLY_COLOR_ANSI_BRIGHT_BGS_EXTRA_MAP


"""

# ANSI constants (copied from evennia.utils.ansi to avoid import)

_ANSI_BEEP = "\07"
_ANSI_ESCAPE = "\033"
_ANSI_NORMAL = "\033[0m"

_ANSI_UNDERLINE = "\033[4m"
_ANSI_HILITE = "\033[1m"
_ANSI_UNHILITE = "\033[22m"
_ANSI_BLINK = "\033[5m"
_ANSI_INVERSE = "\033[7m"
_ANSI_INV_HILITE = "\033[1;7m"
_ANSI_INV_BLINK = "\033[7;5m"
_ANSI_BLINK_HILITE = "\033[1;5m"
_ANSI_INV_BLINK_HILITE = "\033[1;5;7m"

# Foreground colors
_ANSI_BLACK = "\033[30m"
_ANSI_RED = "\033[31m"
_ANSI_GREEN = "\033[32m"
_ANSI_YELLOW = "\033[33m"
_ANSI_BLUE = "\033[34m"
_ANSI_MAGENTA = "\033[35m"
_ANSI_CYAN = "\033[36m"
_ANSI_WHITE = "\033[37m"

# Background colors
_ANSI_BACK_BLACK = "\033[40m"
_ANSI_BACK_RED = "\033[41m"
_ANSI_BACK_GREEN = "\033[42m"
_ANSI_BACK_YELLOW = "\033[43m"
_ANSI_BACK_BLUE = "\033[44m"
_ANSI_BACK_MAGENTA = "\033[45m"
_ANSI_BACK_CYAN = "\033[46m"
_ANSI_BACK_WHITE = "\033[47m"

# Formatting Characters
_ANSI_RETURN = "\r\n"
_ANSI_TAB = "\t"
_ANSI_SPACE = " "


#############################################################
#
# %c - MUX/MUSH style markup. This was Evennia's first
# color markup style. It was phased out due to % being used
# in Python formatting operations.
#
# %ch%cr, %cr - bright/dark red foreground
# %ch%cR, %cR- bright/dark red background
# %c500, %c[500 - XTERM256 red foreground/background
# %c=w, %c[=w - XTERM256 greyscale foreground/background
#
#############################################################

MUX_COLOR_ANSI_EXTRA_MAP = [
    (r"%cn", _ANSI_NORMAL),  # reset
    (r"%ch", _ANSI_HILITE),  # highlight
    (r"%r", _ANSI_RETURN),  # line break
    (r"%R", _ANSI_RETURN),  #
    (r"%t", _ANSI_TAB),  # tab
    (r"%T", _ANSI_TAB),  #
    (r"%b", _ANSI_SPACE),  # space
    (r"%B", _ANSI_SPACE),
    (r"%cf", _ANSI_BLINK),  # annoying and not supported by all clients
    (r"%ci", _ANSI_INVERSE),  # invert
    (r"%cr", _ANSI_RED),
    (r"%cg", _ANSI_GREEN),
    (r"%cy", _ANSI_YELLOW),
    (r"%cb", _ANSI_BLUE),
    (r"%cm", _ANSI_MAGENTA),
    (r"%cc", _ANSI_CYAN),
    (r"%cw", _ANSI_WHITE),
    (r"%cx", _ANSI_BLACK),
    (r"%cR", _ANSI_BACK_RED),
    (r"%cG", _ANSI_BACK_GREEN),
    (r"%cY", _ANSI_BACK_YELLOW),
    (r"%cB", _ANSI_BACK_BLUE),
    (r"%cM", _ANSI_BACK_MAGENTA),
    (r"%cC", _ANSI_BACK_CYAN),
    (r"%cW", _ANSI_BACK_WHITE),
    (r"%cX", _ANSI_BACK_BLACK),
]

MUX_COLOR_XTERM256_EXTRA_FG = [r"%c([0-5])([0-5])([0-5])"]  # %c123 - foreground colour
MUX_COLOR_XTERM256_EXTRA_BG = [r"%c\[([0-5])([0-5])([0-5])"]  # %c[123 - background colour
MUX_COLOR_XTERM256_EXTRA_GFG = [r"%c=([a-z])"]  # %c=a - greyscale foreground
MUX_COLOR_XTERM256_EXTRA_GBG = [r"%c\[=([a-z])"]  # %c[=a - greyscale background

MUX_COLOR_ANSI_XTERM256_BRIGHT_BG_EXTRA_MAP = [
    (r"%ch%cR", r"%c[500"),
    (r"%ch%cG", r"%c[050"),
    (r"%ch%cY", r"%c[550"),
    (r"%ch%cB", r"%c[005"),
    (r"%ch%cM", r"%c[505"),
    (r"%ch%cC", r"%c[055"),
    (r"%ch%cW", r"%c[555"),  # white background
    (r"%ch%cX", r"%c[222"),  # dark grey background
]


#############################################################
#
# {- style MUD markup (old Evennia default). This is
# basically identical to the default |-style except using
# a curly bracket instead. This was removed because {}
# are used in Python string formatting.
#
# WARNING - using this will lead to errors in systems using
# {} mapping (like Object.msg_contents). This is a known error
# that will not be addressed. It's the reason this is no
# longer the default. Use this at your own peril and expect
# to have to fix things yourself.
#
# {r, {R - bright/dark red foreground
# {[r, {[R - bright/dark red background
# {500, {[500 - XTERM256 red foreground/background
# {=w, {[=w - XTERM256 greyscale foreground/background
#
#############################################################

CURLY_COLOR_ANSI_EXTRA_MAP = [
    (r"{n", _ANSI_NORMAL),  # reset
    (r"{/", _ANSI_RETURN),  # line break
    (r"{-", _ANSI_TAB),  # tab
    (r"{_", _ANSI_SPACE),  # space
    (r"{*", _ANSI_INVERSE),  # invert
    (r"{^", _ANSI_BLINK),  # blinking text (very annoying and not supported by all clients)
    (r"{u", _ANSI_UNDERLINE),  # underline
    (r"{r", _ANSI_HILITE + _ANSI_RED),
    (r"{g", _ANSI_HILITE + _ANSI_GREEN),
    (r"{y", _ANSI_HILITE + _ANSI_YELLOW),
    (r"{b", _ANSI_HILITE + _ANSI_BLUE),
    (r"{m", _ANSI_HILITE + _ANSI_MAGENTA),
    (r"{c", _ANSI_HILITE + _ANSI_CYAN),
    (r"{w", _ANSI_HILITE + _ANSI_WHITE),  # pure white
    (r"{x", _ANSI_HILITE + _ANSI_BLACK),  # dark grey
    (r"{R", _ANSI_UNHILITE + _ANSI_RED),
    (r"{G", _ANSI_UNHILITE + _ANSI_GREEN),
    (r"{Y", _ANSI_UNHILITE + _ANSI_YELLOW),
    (r"{B", _ANSI_UNHILITE + _ANSI_BLUE),
    (r"{M", _ANSI_UNHILITE + _ANSI_MAGENTA),
    (r"{C", _ANSI_UNHILITE + _ANSI_CYAN),
    (r"{W", _ANSI_UNHILITE + _ANSI_WHITE),  # light grey
    (r"{X", _ANSI_UNHILITE + _ANSI_BLACK),  # pure black
    # hilight-able colors
    (r"{h", _ANSI_HILITE),
    (r"{H", _ANSI_UNHILITE),
    (r"{!R", _ANSI_RED),
    (r"{!G", _ANSI_GREEN),
    (r"{!Y", _ANSI_YELLOW),
    (r"{!B", _ANSI_BLUE),
    (r"{!M", _ANSI_MAGENTA),
    (r"{!C", _ANSI_CYAN),
    (r"{!W", _ANSI_WHITE),  # light grey
    (r"{!X", _ANSI_BLACK),  # pure black
    # normal ANSI backgrounds
    (r"{[R", _ANSI_BACK_RED),
    (r"{[G", _ANSI_BACK_GREEN),
    (r"{[Y", _ANSI_BACK_YELLOW),
    (r"{[B", _ANSI_BACK_BLUE),
    (r"{[M", _ANSI_BACK_MAGENTA),
    (r"{[C", _ANSI_BACK_CYAN),
    (r"{[W", _ANSI_BACK_WHITE),  # light grey background
    (r"{[X", _ANSI_BACK_BLACK),  # pure black background
]

CURLY_COLOR_XTERM256_EXTRA_FG = [r"\{([0-5])([0-5])([0-5])"]  # |123 - foreground colour
CURLY_COLOR_XTERM256_EXTRA_BG = [r"\{\[([0-5])([0-5])([0-5])"]  # |[123 - background colour
CURLY_COLOR_XTERM256_EXTRA_GFG = [r"\{=([a-z])"]  # |=a - greyscale foreground
CURLY_COLOR_XTERM256_EXTRA_GBG = [r"\{\[=([a-z])"]  # |[=a - greyscale background

CURLY_COLOR_ANSI_XTERM256_BRIGHT_BG_EXTRA_MAP = [
    (r"{[r", r"{[500"),
    (r"{[g", r"{[050"),
    (r"{[y", r"{[550"),
    (r"{[b", r"{[005"),
    (r"{[m", r"{[505"),
    (r"{[c", r"{[055"),
    (r"{[w", r"{[555"),  # white background
    (r"{[x", r"{[222"),  # dark grey background
]



# Generated by Django 3.2.3 on 2021-05-20 21:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("scripts", "0013_auto_20191025_0831"),
    ]

    operations = [
        migrations.AlterField(
            model_name="scriptdb",
            name="db_interval",
            field=models.IntegerField(
                default=-1,
                help_text="how often to repeat script, in seconds. <= 0 means off.",
                verbose_name="interval",
            ),
        ),
        migrations.AlterField(
            model_name="scriptdb",
            name="db_persistent",
            field=models.BooleanField(default=True, verbose_name="survive server reboot"),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2017-01-21 22:30


from django.db import migrations


def update_tags_with_dbmodel(apps, schema_editor):

    ObjectDB = apps.get_model("objects", "ObjectDB")
    for obj in ObjectDB.objects.all():
        obj.db_attributes.all().update(db_model="objectdb")
        obj.db_tags.all().update(db_model="objectdb")

    AccountDB = apps.get_model("accounts", "AccountDB")
    for obj in AccountDB.objects.all():
        obj.db_attributes.all().update(db_model="accountdb")
        obj.db_tags.all().update(db_model="accountdb")

    ScriptDB = apps.get_model("scripts", "ScriptDB")
    for obj in ScriptDB.objects.all():
        obj.db_attributes.all().update(db_model="scriptdb")
        obj.db_tags.all().update(db_model="scriptdb")

    ChannelDB = apps.get_model("comms", "ChannelDB")
    for obj in ChannelDB.objects.all():
        obj.db_attributes.all().update(db_model="channeldb")
        obj.db_tags.all().update(db_model="channeldb")

    HelpEntry = apps.get_model("help", "HelpEntry")
    for obj in HelpEntry.objects.all():
        obj.db_tags.all().update(db_model="helpentry")

    Msg = apps.get_model("comms", "Msg")
    for obj in Msg.objects.all():
        obj.db_tags.all().update(db_model="msg")


class Migration(migrations.Migration):

    dependencies = [("typeclasses", "0005_auto_20160625_1812")]

    operations = [migrations.RunPython(update_tags_with_dbmodel)]



"""
Trace a message through the messaging system
"""

import time


def timetrace(message, idstring, tracemessage="TEST_MESSAGE", final=False):
    """
    Trace a message with time stamps.

    Args:
        message (str): The actual message coming through
        idstring (str): An identifier string specifying where this trace is happening.
        tracemessage (str): The start of the message to tag.
            This message will get attached time stamp.
        final (bool): This is the final leg in the path - include total time in message

    """
    if message.startswith(tracemessage):
        # the message is on the form TEST_MESSAGE tlast t0
        # where t0 is the initial starting time and last is the time
        # saved at the last stop.
        try:
            prefix, tlast, t0 = message.split(None, 2)
            tlast, t0 = float(tlast), float(t0)
        except (IndexError, ValueError):
            t0 = time.time()
            tlast = t0
            t1 = t0
        else:
            t1 = time.time()
        # print to log (important!)
        print("** timetrace (%s): dT=%fs, total=%fs." % (idstring, t1 - tlast, t1 - t0))

        if final:
            message = " ****  %s (total %f) **** " % (tracemessage, t1 - t0)
        else:
            message = "%s %f %f" % (tracemessage, t1, t0)
    return message



"""

Lockfuncs

Lock functions are functions available when defining lock strings,
which in turn limits access to various game systems.

All functions defined globally in this module are assumed to be
available for use in lockstrings to determine access. See the
Evennia documentation for more info on locks.

A lock function is always called with two arguments, accessing_obj and
accessed_obj, followed by any number of arguments. All possible
arguments should be handled with *args, **kwargs. The lock function
should handle all eventual tracebacks by logging the error and
returning False.

Lock functions in this module extend (and will overload same-named)
lock functions from evennia.locks.lockfuncs.

"""

# def myfalse(accessing_obj, accessed_obj, *args, **kwargs):
#    """
#    called in lockstring with myfalse().
#    A simple logger that always returns false. Prints to stdout
#    for simplicity, should use utils.logger for real operation.
#    """
#    print "%s tried to access %s. Access denied." % (accessing_obj, accessed_obj)
#    return False



"""
Evennia Talkative NPC

Contribution - Griatch 2011, grungies1138, 2016

This is a static NPC object capable of holding a simple menu-driven
conversation. It's just meant as an example.

Installation:

Create the NPC by creating an object of typeclass `contrib.tutorials.talking_npc.TalkingNPC`,
For example:

    create/drop John : contrib.tutorials.talking_npc.TalkingNPC

Use `talk` in the same room as the NPC to start a conversation.

If there are many talkative npcs in the same room you will get to
choose which one's talk command to call (Evennia handles this
automatically). This use of EvMenu is very simplistic; See EvMenu for
a lot more complex possibilities.


"""

from evennia import CmdSet, DefaultObject, default_cmds
from evennia.utils.evmenu import EvMenu

# Menu implementing the dialogue tree


def menu_start_node(caller):
    text = "'Hello there, how can I help you?'"

    options = (
        {"desc": "Hey, do you know what this 'Evennia' thing is all about?", "goto": "info1"},
        {"desc": "What's your name, little NPC?", "goto": "info2"},
    )

    return text, options


def info1(caller):
    text = "'Oh, Evennia is where you are right now! Don't you feel the power?'"

    options = (
        {"desc": "Sure, *I* do, not sure how you do though. You are just an NPC.", "goto": "info3"},
        {"desc": "Sure I do. What's yer name, NPC?", "goto": "info2"},
        {"desc": "Ok, bye for now then.", "goto": "END"},
    )

    return text, options


def info2(caller):
    text = "'My name is not really important ... I'm just an NPC after all.'"

    options = (
        {"desc": "I didn't really want to know it anyhow.", "goto": "info3"},
        {"desc": "Okay then, so what's this 'Evennia' thing about?", "goto": "info1"},
    )

    return text, options


def info3(caller):
    text = (
        "'Well ... I'm sort of busy so, have to go. NPC business. "
        "Important stuff. You wouldn't understand.'"
    )

    options = (
        {"desc": "Oookay ... I won't keep you. Bye.", "goto": "END"},
        {"desc": "Wait, why don't you tell me your name first?", "goto": "info2"},
    )

    return text, options


def END(caller):
    text = "'Goodbye, then.'"

    options = ()

    return text, options


#
# The talk command (sits on the NPC)
#


class CmdTalk(default_cmds.MuxCommand):
    """
    Talks to an npc

    Usage:
      talk

    This command is only available if a talkative non-player-character
    (NPC) is actually present. It will strike up a conversation with
    that NPC and give you options on what to talk about.
    """

    key = "talk"
    locks = "cmd:all()"
    help_category = "General"

    def func(self):
        "Implements the command."

        # self.obj is the NPC this is defined on
        self.caller.msg("(You walk up and talk to %s.)" % self.obj.key)

        # Initiate the menu. Change this if you are putting this on
        # some other custom NPC class.
        EvMenu(
            self.caller,
            "evennia.contrib.tutorials.talking_npc.talking_npc",
            startnode="menu_start_node",
        )


class TalkingCmdSet(CmdSet):
    "Stores the talk command."
    key = "talkingcmdset"

    def at_cmdset_creation(self):
        "populates the cmdset"
        self.add(CmdTalk())


class TalkingNPC(DefaultObject):
    """
    This implements a simple Object using the talk command and using
    the conversation defined above.
    """

    def at_object_creation(self):
        "This is called when object is first created."
        self.db.desc = "This is a talkative NPC."
        # assign the talk command to npc
        self.cmdset.add_default(TalkingCmdSet, persistent=True)



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("scripts", "0006_auto_20150310_2249")]

    operations = [
        migrations.DeleteModel(name="DefaultScript"),
        migrations.DeleteModel(name="DoNothing"),
        migrations.DeleteModel(name="ScriptBase"),
        migrations.DeleteModel(name="Store"),
    ]



"""
This sub-package holds the Scripts system. Scripts are database
entities that can store data both in connection to Objects and Accounts
or globally. They may also have a timer-component to execute various
timed effects.

"""



"""
System commands

These are the default commands called by the system commandhandler
when various exceptions occur. If one of these commands are not
implemented and part of the current cmdset, the engine falls back
to a default solution instead.

Some system commands are shown in this module
as a REFERENCE only (they are not all added to Evennia's
default cmdset since they don't currently do anything differently from the
default backup systems hard-wired in the engine).

Overloading these commands in a cmdset can be used to create
interesting effects. An example is using the NoMatch system command
to implement a line-editor where you don't have to start each
line with a command (if there is no match to a known command,
the line is just added to the editor buffer).
"""

from django.conf import settings

# The command keys the engine is calling
# (the actual names all start with __)
from evennia.commands.cmdhandler import CMD_MULTIMATCH, CMD_NOINPUT, CMD_NOMATCH
from evennia.comms.models import ChannelDB
from evennia.utils import create, utils
from evennia.utils.utils import at_search_result

COMMAND_DEFAULT_CLASS = utils.class_from_module(settings.COMMAND_DEFAULT_CLASS)

# Command called when there is no input at line
# (i.e. an lone return key)


class SystemNoInput(COMMAND_DEFAULT_CLASS):
    """
    This is called when there is no input given
    """

    key = CMD_NOINPUT
    locks = "cmd:all()"

    def func(self):
        "Do nothing."
        pass


#
# Command called when there was no match to the
# command name
#
class SystemNoMatch(COMMAND_DEFAULT_CLASS):
    """
    No command was found matching the given input.
    """

    key = CMD_NOMATCH
    locks = "cmd:all()"

    def func(self):
        """
        This is given the failed raw string as input.
        """
        self.msg("Huh?")


#
# Command called when there were multiple matches to the command.
#
class SystemMultimatch(COMMAND_DEFAULT_CLASS):
    """
    Multiple command matches.

    The cmdhandler adds a special attribute 'matches' to this
    system command.

      matches = [(cmdname, args, cmdobj, cmdlen, mratio, raw_cmdname) , (cmdname, ...), ...]

    Here, `cmdname` is the command's name and `args` the rest of the incoming string,
    without said command name. `cmdobj` is the Command instance, the cmdlen is
    the same as len(cmdname) and mratio is a measure of how big a part of the
    full input string the cmdname takes up - an exact match would be 1.0. Finally,
    the `raw_cmdname` is the cmdname unmodified by eventual prefix-stripping.

    """

    key = CMD_MULTIMATCH
    locks = "cmd:all()"

    def func(self):
        """
        Handle multiple-matches by using the at_search_result default handler.

        """
        # this was set by the cmdparser and is a tuple
        #    (cmdname, args, cmdobj, cmdlen, mratio, raw_cmdname). See
        # evennia.commands.cmdparse.create_match for more details.
        matches = self.matches
        # at_search_result will itself msg the multimatch options to the caller.
        at_search_result([match[2] for match in matches], self.caller, query=matches[0][0])



import os
import sys
import time
from os.path import abspath, dirname

from django.conf import settings
from django.db import connection
from twisted.application import internet, service
from twisted.application.service import MultiService
from twisted.internet import protocol, reactor
from twisted.internet.task import LoopingCall

import evennia
from evennia.utils.utils import (
    class_from_module,
    get_evennia_version,
    make_iter,
    mod_import,
)


class EvenniaPortalService(MultiService):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.amp_protocol = None
        self.server_process_id = None
        self.server_restart_mode = "shutdown"
        self.server_info_dict = dict()
        self.plugins = list()

        self.start_time = 0
        self._maintenance_count = 0
        self.maintenance_task = None

        self.info_dict = {
            "servername": settings.SERVERNAME,
            "version": get_evennia_version(),
            "errors": "",
            "info": "",
            "lockdown_mode": "",
            "amp": "",
            "telnet": [],
            "telnet_ssl": [],
            "ssh": [],
            "webclient": [],
            "webserver_proxy": [],
            "webserver_internal": [],
        }

        # in non-interactive portal mode, this gets overwritten by
        # cmdline sent by the evennia launcher
        self.server_twistd_cmd = self._get_backup_server_twistd_cmd()

    def portal_maintenance(self):
        """
        Repeated maintenance tasks for the portal.

        """

        self._maintenance_count += 1

        if self._maintenance_count % (60 * 7) == 0:
            # drop database connection every 7 hrs to avoid default timeouts on MySQL
            # (see https://github.com/evennia/evennia/issues/1376)
            connection.close()

    def privilegedStartService(self):
        self.start_time = time.time()
        self.maintenance_task = LoopingCall(self.portal_maintenance)
        self.maintenance_task.start(60, now=True)  # call every minute
        # set a callback if the server is killed abruptly,
        # by Ctrl-C, reboot etc.
        reactor.addSystemEventTrigger(
            "before", "shutdown", self.shutdown, _reactor_stopping=True, _stop_server=True
        )

        if settings.AMP_HOST and settings.AMP_PORT and settings.AMP_INTERFACE:
            self.register_amp()

        if settings.TELNET_ENABLED and settings.TELNET_PORTS and settings.TELNET_INTERFACES:
            self.register_telnet()

        if settings.SSL_ENABLED and settings.SSL_PORTS and settings.SSL_INTERFACES:
            self.register_ssl()

        if settings.SSH_ENABLED and settings.SSH_PORTS and settings.SSH_INTERFACES:
            self.register_ssh()

        if settings.WEBSERVER_ENABLED:
            self.register_webserver()

        if settings.LOCKDOWN_MODE:
            self.info_dict["lockdown_mode"] = "  LOCKDOWN_MODE active: Only local connections."

        self.register_plugins()

        super().privilegedStartService()

    def register_plugins(self):
        self.plugins.extend(
            mod_import(module) for module in make_iter(settings.PORTAL_SERVICES_PLUGIN_MODULES)
        )
        for plugin_module in self.plugins:
            # external plugin services to start
            if plugin_module:
                plugin_module.start_plugin_services(self)

    def check_lockdown(self, interfaces: list[str]):
        if settings.LOCKDOWN_MODE:
            return ["127.0.0.1"]
        return interfaces

    def register_ssl(self):
        # Start Telnet+SSL game connection (requires PyOpenSSL).

        from evennia.server.portal import telnet_ssl

        _ssl_protocol = class_from_module(settings.SSL_PROTOCOL_CLASS)

        interfaces = self.check_lockdown(settings.SSL_INTERFACES)

        for interface in interfaces:
            ifacestr = ""
            if interface not in ("0.0.0.0", "::") or len(interfaces) > 1:
                ifacestr = "-%s" % interface
            for port in settings.SSL_PORTS:
                pstring = "%s:%s" % (ifacestr, port)
                factory = protocol.ServerFactory()
                factory.noisy = False
                factory.sessionhandler = evennia.PORTAL_SESSION_HANDLER
                factory.protocol = _ssl_protocol

                ssl_context = telnet_ssl.getSSLContext()
                if ssl_context:
                    ssl_service = internet.SSLServer(
                        port, factory, telnet_ssl.getSSLContext(), interface=interface
                    )
                    ssl_service.setName("EvenniaSSL%s" % pstring)
                    ssl_service.setServiceParent(self)

                    self.info_dict["telnet_ssl"].append("telnet+ssl%s: %s" % (ifacestr, port))
                else:
                    self.info_dict["telnet_ssl"].append(
                        "telnet+ssl%s: %s (deactivated - keys/cert unset)" % (ifacestr, port)
                    )

    def register_ssh(self):
        # Start SSH game connections. Will create a keypair in
        # evennia/game if necessary.

        from evennia.server.portal import ssh

        _ssh_protocol = class_from_module(settings.SSH_PROTOCOL_CLASS)

        interfaces = self.check_lockdown(settings.SSH_INTERFACES)

        for interface in interfaces:
            ifacestr = ""
            if interface not in ("0.0.0.0", "::") or len(interfaces) > 1:
                ifacestr = "-%s" % interface
            for port in settings.SSH_PORTS:
                pstring = "%s:%s" % (ifacestr, port)
                factory = ssh.makeFactory(
                    {
                        "protocolFactory": _ssh_protocol,
                        "protocolArgs": (),
                        "sessions": evennia.PORTAL_SESSION_HANDLER,
                    }
                )
                factory.noisy = False
                ssh_service = internet.TCPServer(port, factory, interface=interface)
                ssh_service.setName("EvenniaSSH%s" % pstring)
                ssh_service.setServiceParent(self)

                self.info_dict["ssh"].append("ssh%s: %s" % (ifacestr, port))

    def register_webserver(self):
        from evennia.server.webserver import EvenniaReverseProxyResource, Website

        # Start a reverse proxy to relay data to the Server-side webserver
        interfaces = self.check_lockdown(settings.WEBSERVER_INTERFACES)
        websocket_started = False
        _websocket_protocol = class_from_module(settings.WEBSOCKET_PROTOCOL_CLASS)
        for interface in interfaces:
            ifacestr = ""
            if interface not in ("0.0.0.0", "::") or len(interfaces) > 1:
                ifacestr = "-%s" % interface

            for proxyport, serverport in settings.WEBSERVER_PORTS:
                web_root = EvenniaReverseProxyResource("127.0.0.1", serverport, "")
                webclientstr = ""
                if settings.WEBCLIENT_ENABLED:
                    # create ajax client processes at /webclientdata
                    ajax_class = class_from_module(settings.AJAX_CLIENT_CLASS)
                    ajax_webclient = ajax_class()
                    ajax_webclient.sessionhandler = evennia.PORTAL_SESSION_HANDLER
                    web_root.putChild(b"webclientdata", ajax_webclient)
                    webclientstr = "webclient (ajax only)"

                    if (
                        settings.WEBSOCKET_CLIENT_ENABLED
                        and settings.WEBSOCKET_CLIENT_PORT
                        and settings.WEBSOCKET_CLIENT_INTERFACE
                    ) and not websocket_started:
                        # start websocket client port for the webclient
                        # we only support one websocket client
                        from autobahn.twisted.websocket import WebSocketServerFactory

                        from evennia.server.portal import webclient  # noqa

                        w_interface = (
                            "127.0.0.1"
                            if settings.LOCKDOWN_MODE
                            else settings.WEBSOCKET_CLIENT_INTERFACE
                        )
                        w_ifacestr = ""
                        if (
                            w_interface not in ("0.0.0.0", "::")
                            or len(settings.WEBSERVER_INTERFACES) > 1
                        ):
                            w_ifacestr = "-%s" % w_interface
                        port = settings.WEBSOCKET_CLIENT_PORT

                        class Websocket(WebSocketServerFactory):
                            "Only here for better naming in logs"
                            pass

                        factory = Websocket()
                        factory.noisy = False
                        factory.protocol = _websocket_protocol
                        factory.sessionhandler = evennia.PORTAL_SESSION_HANDLER
                        websocket_service = internet.TCPServer(port, factory, interface=w_interface)
                        websocket_service.setName("EvenniaWebSocket%s:%s" % (w_ifacestr, port))
                        websocket_service.setServiceParent(self)
                        websocket_started = True
                        webclientstr = "webclient-websocket%s: %s" % (w_ifacestr, port)
                    self.info_dict["webclient"].append(webclientstr)

                try:
                    WEB_PLUGINS_MODULE = mod_import(settings.WEB_PLUGINS_MODULE)
                except ImportError:
                    WEB_PLUGINS_MODULE = None
                    self.info_dict["errors"] = (
                        "WARNING: settings.WEB_PLUGINS_MODULE not found - "
                        "copy 'evennia/game_template/server/conf/web_plugins.py to "
                        "mygame/server/conf."
                    )

                if WEB_PLUGINS_MODULE:
                    try:
                        web_root = WEB_PLUGINS_MODULE.at_webproxy_root_creation(web_root)
                    except Exception:
                        # Legacy user has not added an at_webproxy_root_creation function in existing
                        # web plugins file
                        self.info_dict["errors"] = (
                            "WARNING: WEB_PLUGINS_MODULE is enabled but at_webproxy_root_creation() "
                            "not found copy 'evennia/game_template/server/conf/web_plugins.py to "
                            "mygame/server/conf."
                        )
                web_root = Website(web_root, logPath=settings.HTTP_LOG_FILE)
                web_root.is_portal = True
                proxy_service = internet.TCPServer(proxyport, web_root, interface=interface)
                proxy_service.setName("EvenniaWebProxy%s:%s" % (ifacestr, proxyport))
                proxy_service.setServiceParent(self)
                self.info_dict["webserver_proxy"].append(
                    "webserver-proxy%s: %s" % (ifacestr, proxyport)
                )
                self.info_dict["webserver_internal"].append("webserver: %s" % serverport)

    def register_telnet(self):
        # Start telnet game connections

        from evennia.server.portal import telnet

        _telnet_protocol = class_from_module(settings.TELNET_PROTOCOL_CLASS)

        interfaces = self.check_lockdown(settings.TELNET_INTERFACES)

        for interface in interfaces:
            ifacestr = ""
            if interface not in ("0.0.0.0", "::") or len(interfaces) > 1:
                ifacestr = "-%s" % interface
            for port in settings.TELNET_PORTS:
                pstring = "%s:%s" % (ifacestr, port)
                factory = telnet.TelnetServerFactory()
                factory.noisy = False
                factory.protocol = _telnet_protocol
                factory.sessionhandler = evennia.PORTAL_SESSION_HANDLER
                telnet_service = internet.TCPServer(port, factory, interface=interface)
                telnet_service.setName("EvenniaTelnet%s" % pstring)
                telnet_service.setServiceParent(self)

                self.info_dict["telnet"].append("telnet%s: %s" % (ifacestr, port))

    def register_amp(self):
        # The AMP protocol handles the communication between
        # the portal and the mud server. Only reason to ever deactivate
        # it would be during testing and debugging.

        from evennia.server.portal import amp_server

        self.info_dict["amp"] = "amp: %s" % settings.AMP_PORT

        factory = amp_server.AMPServerFactory(self)
        amp_service = internet.TCPServer(
            settings.AMP_PORT, factory, interface=settings.AMP_INTERFACE
        )
        amp_service.setName("PortalAMPServer")
        amp_service.setServiceParent(self)

    def _get_backup_server_twistd_cmd(self):
        """
        For interactive Portal mode there is no way to get the server cmdline from the launcher, so
        we need to guess it here (it's very likely to not change)

        Returns:
            server_twistd_cmd (list): An instruction for starting the server, to pass to Popen.

        """
        server_twistd_cmd = [
            "twistd",
            "--python={}".format(os.path.join(dirname(dirname(abspath(__file__))), "server.py")),
        ]
        if os.name != "nt":
            gamedir = os.getcwd()
            server_twistd_cmd.append(
                "--pidfile={}".format(os.path.join(gamedir, "server", "server.pid"))
            )
        return server_twistd_cmd

    def get_info_dict(self):
        """
        Return the Portal info, for display.

        """
        return self.info_dict

    def shutdown(self, _reactor_stopping=False, _stop_server=False):
        """
        Shuts down the server from inside it.

        Args:
            _reactor_stopping (bool, optional): This is set if server
                is already in the process of shutting down; in this case
                we don't need to stop it again.
            _stop_server (bool, optional): Only used in portal-interactive mode;
                makes sure to stop the Server cleanly.

        Note that restarting (regardless of the setting) will not work
        if the Portal is currently running in daemon mode. In that
        case it always needs to be restarted manually.

        """
        if _reactor_stopping and hasattr(self, "shutdown_complete"):
            # we get here due to us calling reactor.stop below. No need
            # to do the shutdown procedure again.
            return

        evennia.PORTAL_SESSION_HANDLER.disconnect_all()
        if _stop_server:
            self.amp_protocol.stop_server(mode="shutdown")
        if not _reactor_stopping:
            # shutting down the reactor will trigger another signal. We set
            # a flag to avoid loops.
            self.shutdown_complete = True
            reactor.callLater(0, reactor.stop)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-07-09 21:33


import re

from django.db import migrations

CASE_WORD_EXCEPTIONS = []


def _case_sensitive_replace(string, old, new):
    """
    Replace text, retaining exact case.

    Args:
        string (str): String in which to perform replacement.
        old (str): Word or substring to replace.
        new (str): What to replace `old` with.

    Returns:
        repl_string (str): Version of string where instances of
            `old` has been replaced with `new`, retaining case.

    """

    def repl(match):
        current = match.group()
        # treat multi-word sentences word-by-word
        old_words = current.split(" ")
        new_words = new.split(" ")
        out = []
        for old_word, new_word in zip(old_words, new_words):
            result = []
            all_upper = True
            for ind, chr in enumerate(old_word):
                if ind >= len(new):
                    break
                if chr.isupper():
                    result.append(new_word[ind].upper())
                else:
                    result.append(new_word[ind].lower())
                    all_upper = False
            # special cases - keep remaing case)
            if new_word.lower() in CASE_WORD_EXCEPTIONS:
                result.append(new_word[ind + 1 :])
            # append any remaining characters from new
            elif all_upper:
                result.append(new_word[ind + 1 :].upper())
            else:
                result.append(new_word[ind + 1 :].lower())
            out.append("".join(result))
        # if we have more new words than old ones, just add them verbatim
        out.extend([new_word for ind, new_word in enumerate(new_words) if ind >= len(old_words)])
        return " ".join(out)

    if string is None:
        return None

    regex = re.compile(re.escape(old), re.I)
    return regex.sub(repl, string)


def update_typeclasses(apps, schema_editor):

    ObjectDB = apps.get_model("objects", "ObjectDB")
    AccountDB = apps.get_model("accounts", "AccountDB")
    ScriptDB = apps.get_model("scripts", "ScriptDB")
    ChannelDB = apps.get_model("comms", "ChannelDB")
    Attributes = apps.get_model("typeclasses", "Attribute")
    Tags = apps.get_model("typeclasses", "Tag")

    for obj in ObjectDB.objects.all():
        obj.db_typeclass_path = _case_sensitive_replace(obj.db_typeclass_path, "player", "account")
        obj.db_cmdset_storage = _case_sensitive_replace(obj.db_cmdset_storage, "player", "account")
        obj.db_lock_storage = _case_sensitive_replace(obj.db_lock_storage, "player", "account")
        obj.save(update_fields=["db_typeclass_path", "db_cmdset_storage", "db_lock_storage"])
    for obj in AccountDB.objects.all():
        obj.db_typeclass_path = _case_sensitive_replace(obj.db_typeclass_path, "player", "account")
        obj.db_cmdset_storage = _case_sensitive_replace(obj.db_cmdset_storage, "player", "account")
        obj.db_lock_storage = _case_sensitive_replace(obj.db_lock_storage, "player", "account")
        obj.save(update_fields=["db_typeclass_path", "db_cmdset_storage", "db_lock_storage"])
    for obj in ScriptDB.objects.all():
        obj.db_typeclass_path = _case_sensitive_replace(obj.db_typeclass_path, "player", "account")
        obj.db_lock_storage = _case_sensitive_replace(obj.db_lock_storage, "player", "account")
        obj.save(update_fields=["db_typeclass_path", "db_lock_storage"])
    for obj in ChannelDB.objects.all():
        obj.db_typeclass_path = _case_sensitive_replace(obj.db_typeclass_path, "player", "account")
        obj.db_lock_storage = _case_sensitive_replace(obj.db_lock_storage, "player", "account")
        obj.save(update_fields=["db_typeclass_path", "db_lock_storage"])
    for obj in Attributes.objects.filter(db_model="playerdb"):
        obj.db_model = "accountdb"
        obj.save(update_fields=["db_model"])
    for obj in Tags.objects.filter(db_model="playerdb"):
        obj.db_model = "accountdb"
        obj.save(update_fields=["db_model"])


class Migration(migrations.Migration):

    dependencies = [("typeclasses", "0008_lock_and_perm_rename")]

    operations = [migrations.RunPython(update_typeclasses)]



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("scripts", "0004_auto_20150306_1354")]

    operations = [
        migrations.DeleteModel(name="CheckSessions"),
        migrations.DeleteModel(name="ValidateChannelHandler"),
        migrations.DeleteModel(name="ValidateIdmapperCache"),
        migrations.DeleteModel(name="ValidateScripts"),
    ]



"""
Outgoing callables to apply with the FuncParser on outgoing messages.

The functions in this module will become available as $funcname(args, kwargs)
in all outgoing strings if you add

    FUNCPARSER_PARSE_OUTGOING_MESSAGES_ENABLED = True

to your settings file. The default inlinefuncs are found at the bottom of
`evennia.utils.funcparser`.

In text, usage is straightforward:

$funcname(arg1, arg2, ..., key=val, key2=val2, ...)

Example 1 (using the "pad" inlinefunc):
    say This is $pad("a center-padded text", 50,c,-) of width 50.
    ->
    John says, "This is -------------- a center-padded text--------------- of width 50."

Example 2 (using nested "pad" and "time" inlinefuncs):
    say The time is $pad($time(), 30)right now.
    ->
    John says, "The time is         Oct 25, 11:09         right now."

To add more inline functions, add them to this module, using
the following call signature:

    def funcname(*args, **kwargs)
        ...

"""

# def capitalize(*args, **kwargs):
#    "Silly capitalize example. Used as  $capitalize
#    if not args:
#        return ''
#    session = kwargs.get("session")
#    return args[0].capitalize()



"""

MCCP - Mud Client Compression Protocol

This implements the MCCP v2 telnet protocol as per
http://tintin.sourceforge.net/mccp/. MCCP allows for the server to
compress data when sending to supporting clients, reducing bandwidth
by 70-90%.. The compression is done using Python's builtin zlib
library. If the client doesn't support MCCP, server sends uncompressed
as normal.  Note: On modern hardware you are not likely to notice the
effect of MCCP unless you have extremely heavy traffic or sits on a
terribly slow connection.

This protocol is implemented by the telnet protocol importing
mccp_compress and calling it from its write methods.
"""
import zlib

# negotiations for v1 and v2 of the protocol
MCCP = bytes([86])  # b"\x56"
FLUSH = zlib.Z_SYNC_FLUSH


def mccp_compress(protocol, data):
    """
    Handles zlib compression, if applicable.

    Args:
        data (str): Incoming data to compress.

    Returns:
        stream (binary): Zlib-compressed data.

    """
    if hasattr(protocol, "zlib"):
        return protocol.zlib.compress(data) + protocol.zlib.flush(FLUSH)
    return data


class Mccp:
    """
    Implements the MCCP protocol. Add this to a
    variable on the telnet protocol to set it up.

    """

    def __init__(self, protocol):
        """
        initialize MCCP by storing protocol on
        ourselves and calling the client to see if
        it supports MCCP. Sets callbacks to
        start zlib compression in that case.

        Args:
            protocol (Protocol): The active protocol instance.

        """

        self.protocol = protocol
        self.protocol.protocol_flags["MCCP"] = False
        # ask if client will mccp, connect callbacks to handle answer
        self.protocol.will(MCCP).addCallbacks(self.do_mccp, self.no_mccp)

    def no_mccp(self, option):
        """
        Called if client doesn't support mccp or chooses to turn it off.

        Args:
            option (Option): Option dict (not used).

        """
        if hasattr(self.protocol, "zlib"):
            del self.protocol.zlib
        self.protocol.protocol_flags["MCCP"] = False
        self.protocol.handshake_done()

    def do_mccp(self, option):
        """
        The client supports MCCP. Set things up by
        creating a zlib compression stream.

        Args:
            option (Option): Option dict (not used).

        """
        self.protocol.protocol_flags["MCCP"] = True
        self.protocol.requestNegotiation(MCCP, b"")
        self.protocol.zlib = zlib.compressobj(9)
        self.protocol.handshake_done()



"""
This module implements the main Evennia server process, the core of the game
engine.

This module should be started with the 'twistd' executable since it sets up all
the networking features.  (this is done automatically by
evennia/server/server_runner.py).

"""
import os
import sys

import django
from twisted.logger import globalLogPublisher

django.setup()

import evennia

evennia._init()

from django.conf import settings

from evennia.utils import logger

# twistd requires us to define the variable 'application' so it knows
# what to execute from.
# The guts of the application are in the service.py file,
# which is instantiated and attached to application in evennia._init()
application = evennia.TWISTED_APPLICATION

if "--nodaemon" not in sys.argv and "test" not in sys.argv:
    # activate logging for interactive/testing mode
    logfile = logger.WeeklyLogFile(
        os.path.basename(settings.SERVER_LOG_FILE),
        os.path.dirname(settings.SERVER_LOG_FILE),
        day_rotation=settings.SERVER_LOG_DAY_ROTATION,
        max_size=settings.SERVER_LOG_MAX_SIZE,
    )
    globalLogPublisher.addObserver(logger.GetServerLogObserver()(logfile))



from django.contrib.auth.models import AnonymousUser
from django.test import RequestFactory, TestCase
from mock import MagicMock, patch

from . import general_context


class TestGeneralContext(TestCase):
    maxDiff = None

    @patch("evennia.web.utils.general_context.GAME_NAME", "test_name")
    @patch("evennia.web.utils.general_context.GAME_SLOGAN", "test_game_slogan")
    @patch("evennia.web.utils.general_context.REGISTER_ENABLED", "register_enabled_testvalue")
    @patch(
        "evennia.web.utils.general_context.WEBSOCKET_CLIENT_ENABLED",
        "websocket_client_enabled_testvalue",
    )
    @patch("evennia.web.utils.general_context.WEBCLIENT_ENABLED", "webclient_enabled_testvalue")
    @patch("evennia.web.utils.general_context.WEBSOCKET_PORT", "websocket_client_port_testvalue")
    @patch("evennia.web.utils.general_context.WEBSOCKET_URL", "websocket_client_url_testvalue")
    @patch("evennia.web.utils.general_context.REST_API_ENABLED", True)
    def test_general_context(self):
        request = RequestFactory().get("/")
        request.user = AnonymousUser()
        request.session = {"account": None, "puppet": None}

        response = general_context.general_context(request)

        self.assertEqual(
            response,
            {
                "account": None,
                "puppet": None,
                "game_name": "test_name",
                "game_slogan": "test_game_slogan",
                "evennia_userapps": ["Accounts"],
                "evennia_entityapps": ["Objects", "Scripts", "Comms", "Help"],
                "evennia_setupapps": ["Permissions", "Config"],
                "evennia_connectapps": ["Irc"],
                "evennia_websiteapps": ["Flatpages", "News", "Sites"],
                "register_enabled": "register_enabled_testvalue",
                "webclient_enabled": "webclient_enabled_testvalue",
                "websocket_enabled": "websocket_client_enabled_testvalue",
                "websocket_port": "websocket_client_port_testvalue",
                "websocket_url": "websocket_client_url_testvalue",
                "rest_api_enabled": True,
                "server_hostname": "localhost",
                "ssh_enabled": False,
                "ssh_ports": False,
                "telnet_enabled": True,
                "telnet_ports": [4000],
                "telnet_ssl_enabled": False,
                "telnet_ssl_ports": [4003],
            },
        )



"""
Godot Websocket - ChrisLR 2022

This provides parsing the ansi text to bbcode used by Godot for their RichTextLabel
and also provides the proper portal service to dedicate a port for Godot's Websockets.

This allows you to connect both the regular webclient and a godot specific webclient.
You can simply connect the resulting text to Godot's RichTextLabel and have the proper display.
You could also pass extra data to this client for advanced functionality.

See the docs for more information.
"""
from evennia.contrib.base_systems.godotwebsocket.text2bbcode import (
    BBCODE_PARSER,
    parse_to_bbcode,
)
from evennia.contrib.base_systems.godotwebsocket.webclient import GodotWebSocketClient



"""
Various helper resources for writing unittests.

Classes for testing Evennia core:

- `BaseEvenniaTestCase` - no default objects, only enforced default settings
- `BaseEvenniaTest` - all default objects, enforced default settings
- `BaseEvenniaCommandTest` - for testing Commands, enforced default settings

Classes for testing game folder content:

- `EvenniaTestCase` - no default objects, using gamedir settings (identical to
   standard Python TestCase)
- `EvenniaTest` - all default objects, using gamedir settings
- `EvenniaCommandTest` - for testing game folder commands, using gamedir settings

Other:

- `EvenniaTestMixin` - A class mixin for creating the test environment objects, for
  making custom tests.
- `EvenniaCommandMixin` - A class mixin that adds support for command testing with the .call()
  helper. Used by the command-test classes, but can be used for making a customt test class.

"""
import re
import sys
import types

from django.conf import settings
from django.test import TestCase, override_settings
from mock import MagicMock, Mock, patch
from twisted.internet.defer import Deferred

import evennia
from evennia import settings_default
from evennia.accounts.accounts import DefaultAccount
from evennia.commands.command import InterruptCommand
from evennia.commands.default.muxcommand import MuxCommand
from evennia.objects.objects import (
    DefaultCharacter,
    DefaultExit,
    DefaultObject,
    DefaultRoom,
)
from evennia.scripts.scripts import DefaultScript
from evennia.server.serversession import ServerSession
from evennia.utils import ansi, create
from evennia.utils.idmapper.models import flush_cache
from evennia.utils.utils import all_from_module, to_str

_RE_STRIP_EVMENU = re.compile(r"^\+|-+\+|\+-+|--+|\|(?:\s|$)", re.MULTILINE)


# set up a 'pristine' setting, unaffected by any changes in mygame
DEFAULT_SETTING_RESETS = dict(
    CONNECTION_SCREEN_MODULE="evennia.game_template.server.conf.connection_screens",
    AT_SERVER_STARTSTOP_MODULE="evennia.game_template.server.conf.at_server_startstop",
    AT_SERVICES_PLUGINS_MODULES=["evennia.game_template.server.conf.server_services_plugins"],
    PORTAL_SERVICES_PLUGIN_MODULES=["evennia.game_template.server.conf.portal_services_plugins"],
    MSSP_META_MODULE="evennia.game_template.server.conf.mssp",
    WEB_PLUGINS_MODULE="server.conf.web_plugins",
    LOCK_FUNC_MODULES=("evennia.locks.lockfuncs", "evennia.game_template.server.conf.lockfuncs"),
    INPUT_FUNC_MODULES=[
        "evennia.server.inputfuncs",
        "evennia.game_template.server.conf.inputfuncs",
    ],
    PROTOTYPE_MODULES=["evennia.game_template.world.prototypes"],
    CMDSET_UNLOGGEDIN="evennia.game_template.commands.default_cmdsets.UnloggedinCmdSet",
    CMDSET_SESSION="evennia.game_template.commands.default_cmdsets.SessionCmdSet",
    CMDSET_CHARACTER="evennia.game_template.commands.default_cmdsets.CharacterCmdSet",
    CMDSET_ACCOUNT="evennia.game_template.commands.default_cmdsets.AccountCmdSet",
    CMDSET_PATHS=["evennia.game_template.commands", "evennia", "evennia.contrib"],
    TYPECLASS_PATHS=[
        "evennia",
        "evennia.contrib",
        "evennia.contrib.game_systems",
        "evennia.contrib.base_systems",
        "evennia.contrib.full_systems",
        "evennia.contrib.tutorials",
        "evennia.contrib.utils",
    ],
    BASE_ACCOUNT_TYPECLASS="evennia.accounts.accounts.DefaultAccount",
    BASE_OBJECT_TYPECLASS="evennia.objects.objects.DefaultObject",
    BASE_CHARACTER_TYPECLASS="evennia.objects.objects.DefaultCharacter",
    BASE_ROOM_TYPECLASS="evennia.objects.objects.DefaultRoom",
    BASE_EXIT_TYPECLASS="evennia.objects.objects.DefaultExit",
    BASE_CHANNEL_TYPECLASS="evennia.comms.comms.DefaultChannel",
    BASE_SCRIPT_TYPECLASS="evennia.scripts.scripts.DefaultScript",
    BASE_BATCHPROCESS_PATHS=[
        "evennia.game_template.world",
        "evennia.contrib",
        "evennia.contrib.tutorials",
    ],
    FILE_HELP_ENTRY_MODULES=["evennia.game_template.world.help_entries"],
    FUNCPARSER_OUTGOING_MESSAGES_MODULES=[
        "evennia.utils.funcparser",
        "evennia.game_template.server.conf.inlinefuncs",
    ],
    FUNCPARSER_PROTOTYPE_PARSING_MODULES=[
        "evennia.prototypes.protfuncs",
        "evennia.game_template.server.conf.prototypefuncs",
    ],
    BASE_GUEST_TYPECLASS="evennia.accounts.accounts.DefaultGuest",
    # a special setting boolean TEST_ENVIRONMENT is set by the test runner
    # while the test suite is running.
    DEFAULT_HOME="#1",
    TEST_ENVIRONMENT=True,
)

DEFAULT_SETTINGS = {**all_from_module(settings_default), **DEFAULT_SETTING_RESETS}
DEFAULT_SETTINGS.pop("DATABASES")  # we want different dbs tested in CI


# mocking of evennia.utils.utils.delay
def mockdelay(timedelay, callback, *args, **kwargs):
    callback(*args, **kwargs)
    return Deferred()


# mocking of twisted's deferLater
def mockdeferLater(reactor, timedelay, callback, *args, **kwargs):
    callback(*args, **kwargs)
    return Deferred()


def unload_module(module):
    """
    Reset import so one can mock global constants.

    Args:
        module (module, object or str): The module will
            be removed so it will have to be imported again. If given
            an object, the module in which that object sits will be unloaded. A string
            should directly give the module pathname to unload.

    Example:

        ```python
        # (in a test method)
        unload_module(foo)
        with mock.patch("foo.GLOBALTHING", "mockval"):
            import foo
            ... # test code using foo.GLOBALTHING, now set to 'mockval'
        ```

    This allows for mocking constants global to the module, since
    otherwise those would not be mocked (since a module is only
    loaded once).

    """
    if isinstance(module, str):
        modulename = module
    elif hasattr(module, "__module__"):
        modulename = module.__module__
    else:
        modulename = module.__name__

    if modulename in sys.modules:
        del sys.modules[modulename]


def _mock_deferlater(reactor, timedelay, callback, *args, **kwargs):
    callback(*args, **kwargs)
    return Deferred()


class EvenniaTestMixin:
    """
    Evennia test environment mixin
    """

    account_typeclass = DefaultAccount
    object_typeclass = DefaultObject
    character_typeclass = DefaultCharacter
    exit_typeclass = DefaultExit
    room_typeclass = DefaultRoom
    script_typeclass = DefaultScript

    def create_accounts(self):
        self.account = create.create_account(
            "TestAccount",
            email="test@test.com",
            password="testpassword",
            typeclass=self.account_typeclass,
        )
        self.account2 = create.create_account(
            "TestAccount2",
            email="test@test.com",
            password="testpassword",
            typeclass=self.account_typeclass,
        )
        self.account.permissions.add("Developer")

    def teardown_accounts(self):
        if hasattr(self, "account"):
            self.account.delete()
        if hasattr(self, "account2"):
            self.account2.delete()

    # Set up fake prototype module for allowing tests to use named prototypes.
    @override_settings(PROTOTYPE_MODULES=["evennia.utils.tests.data.prototypes_example"])
    def create_rooms(self):
        self.room1 = create.create_object(self.room_typeclass, key="Room", nohome=True)
        self.room1.db.desc = "room_desc"

        self.room2 = create.create_object(self.room_typeclass, key="Room2")
        self.exit = create.create_object(
            self.exit_typeclass, key="out", location=self.room1, destination=self.room2
        )

    def create_objs(self):
        self.obj1 = create.create_object(
            self.object_typeclass, key="Obj", location=self.room1, home=self.room1
        )
        self.obj2 = create.create_object(
            self.object_typeclass, key="Obj2", location=self.room1, home=self.room1
        )

    def create_chars(self):
        self.char1 = create.create_object(
            self.character_typeclass, key="Char", location=self.room1, home=self.room1
        )
        self.char1.permissions.add("Developer")
        self.char2 = create.create_object(
            self.character_typeclass, key="Char2", location=self.room1, home=self.room1
        )
        self.char1.account = self.account
        self.account.db._last_puppet = self.char1
        self.char2.account = self.account2
        self.account2.db._last_puppet = self.char2

    def create_script(self):
        self.script = create.create_script(self.script_typeclass, key="Script")

    def setup_session(self):
        dummysession = ServerSession()
        dummysession.init_session("telnet", ("localhost", "testmode"), evennia.SESSION_HANDLER)
        dummysession.sessid = 1
        evennia.SESSION_HANDLER.portal_connect(
            dummysession.get_sync_data()
        )  # note that this creates a new Session!
        session = evennia.SESSION_HANDLER.session_from_sessid(1)  # the real session
        evennia.SESSION_HANDLER.login(session, self.account, testmode=True)
        self.session = session

    def teardown_session(self):
        if hasattr(self, "sessions"):
            del evennia.SESSION_HANDLER[self.session.sessid]

    @patch("evennia.scripts.taskhandler.deferLater", _mock_deferlater)
    def setUp(self):
        """
        Sets up testing environment
        """
        self.backups = (
            evennia.SESSION_HANDLER.data_out,
            evennia.SESSION_HANDLER.disconnect,
            settings.DEFAULT_HOME,
            settings.PROTOTYPE_MODULES,
        )
        evennia.SESSION_HANDLER.data_out = Mock()
        evennia.SESSION_HANDLER.disconnect = Mock()

        self.create_accounts()
        self.create_rooms()
        self.create_objs()
        self.create_chars()
        self.create_script()
        self.setup_session()

    @override_settings(PROTOTYPE_MODULES=["evennia.utils.tests.data.prototypes_example"])
    def tearDown(self):
        flush_cache()
        try:
            evennia.SESSION_HANDLER.data_out = self.backups[0]
            evennia.SESSION_HANDLER.disconnect = self.backups[1]
            settings.DEFAULT_HOME = self.backups[2]
            settings.PROTOTYPE_MODULES = self.backups[3]
        except AttributeError as err:
            raise AttributeError(
                f"{err}: Teardown error. If you overrode the `setUp()` method "
                "in your test, make sure you also added `super().setUp()`!"
            )

        del evennia.SESSION_HANDLER[self.session.sessid]
        self.teardown_accounts()
        super().tearDown()


@patch("evennia.server.portal.portal.LoopingCall", new=MagicMock())
class EvenniaCommandTestMixin:
    """
    Mixin to add to a test in order to provide the `.call` helper for
    testing the execution and returns of a command.

    Tests a Command by running it and comparing what messages it sends with
    expected values. This tests without actually spinning up the cmdhandler
    for every test, which is more controlled.

    Example:
    ::

        from commands.echo import CmdEcho

        class MyCommandTest(EvenniaTest, CommandTestMixin):

            def test_echo(self):
                '''
                Test that the echo command really returns
                what you pass into it.
                '''
                self.call(MyCommand(), "hello world!",
                          "You hear your echo: 'Hello world!'")

    """

    # formatting for .call's error message
    _ERROR_FORMAT = """
=========================== Wanted message ===================================
{expected_msg}
=========================== Returned message =================================
{returned_msg}
==============================================================================
""".rstrip()

    def call(
        self,
        cmdobj,
        input_args,
        msg=None,
        cmdset=None,
        noansi=True,
        caller=None,
        receiver=None,
        cmdstring=None,
        obj=None,
        inputs=None,
        raw_string=None,
    ):
        """
        Test a command by assigning all the needed properties to a cmdobj and
        running the sequence. The resulting `.msg` calls will be mocked and
        the text= calls to them compared to a expected output.

        Args:
            cmdobj (Command): The command object to use.
            input_args (str): This should be the full input the Command should
                see, such as 'look here'. This will become `.args` for the Command
                instance to parse.
            msg (str or dict, optional): This is the expected return value(s)
                returned through `caller.msg(text=...)` calls in the command. If a string, the
                receiver is controlled with the `receiver` kwarg (defaults to `caller`).
                If this is a `dict`, it is a mapping
                `{receiver1: "expected1", receiver2: "expected2",...}` and `receiver` is
                ignored. The message(s) are compared with the actual messages returned
                to the receiver(s) as the Command runs. Each check uses `.startswith`,
                so you can choose to only include the first part of the
                returned message if that's enough to verify a correct result. EvMenu
                decorations (like borders) are stripped and should not be included. This
                should also not include color tags unless `noansi=False`.
                If the command returns texts in multiple separate `.msg`-
                calls to a receiver, separate these with `|` if `noansi=True`
                (default) and `||` if `noansi=False`. If no `msg` is given (`None`),
                then no automatic comparison will be done.
            cmdset (str, optional): If given, make `.cmdset` available on the Command
                instance as it runs. While `.cmdset` is normally available on the
                Command instance by default, this is usually only used by
                commands that explicitly operates/displays cmdsets, like
                `examine`.
            noansi (str, optional): By default the color tags of the `msg` is
                ignored, this makes them significant. If unset, `msg` must contain
                the same color tags as the actual return message.
            caller (Object or Account, optional): By default `self.char1` is used as the
                command-caller (the `.caller` property on the Command). This allows to
                execute with another caller, most commonly an Account.
            receiver (Object or Account, optional): This is the object to receive the
                return messages we want to test. By default this is the same as `caller`
                (which in turn defaults to is `self.char1`). Note that if `msg` is
                a `dict`, this is ignored since the receiver is already specified there.
            cmdstring (str, optional): Normally this is the Command's `key`.
                This allows for tweaking the `.cmdname` property of the
                Command`.  This isb used for commands with multiple aliases,
                where the command explicitly checs which alias was used to
                determine its functionality.
            obj (str, optional): This sets the `.obj` property of the Command - the
                object on which the Command 'sits'. By default this is the same as `caller`.
                This can be used for testing on-object Command interactions.
            inputs (list, optional): A list of strings to pass to functions that pause to
                take input from the user (normally using `@interactive` and
                `ret = yield(question)` or `evmenu.get_input`). Each  element of the
                list will be passed into the command as if the user answered each prompt
                in that order.
            raw_string (str, optional): Normally the `.raw_string` property  is set as
                a combination of your `key/cmdname` and `input_args`. This allows
                direct control of what this is, for example for testing edge cases
                or malformed inputs.

        Returns:
            str or dict: The message sent to `receiver`, or a dict of
                `{receiver: "msg", ...}` if multiple are given. This is usually
                only used with `msg=None` to do the validation externally.

        Raises:
            AssertionError: If the returns of `.msg` calls (tested with `.startswith`) does not
                match `expected_input`.

        Notes:
            As part of the tests, all methods of the Command will be called in
            the proper order:

            - cmdobj.at_pre_cmd()
            - cmdobj.parse()
            - cmdobj.func()
            - cmdobj.at_post_cmd()

        """
        # The `self.char1` is created in the `EvenniaTest` base along with
        # other helper objects like self.room and self.obj
        caller = caller if caller else self.char1
        cmdobj.caller = caller
        cmdobj.cmdname = cmdstring if cmdstring else cmdobj.key
        cmdobj.raw_cmdname = cmdobj.cmdname
        cmdobj.cmdstring = cmdobj.cmdname  # deprecated
        cmdobj.args = input_args
        cmdobj.cmdset = cmdset
        cmdobj.session = evennia.SESSION_HANDLER.session_from_sessid(1)
        cmdobj.account = self.account
        cmdobj.raw_string = raw_string if raw_string is not None else cmdobj.key + " " + input_args
        cmdobj.obj = obj or (caller if caller else self.char1)
        inputs = inputs or []

        # set up receivers
        receiver_mapping = {}
        if isinstance(msg, dict):
            # a mapping {receiver: msg, ...}
            receiver_mapping = {
                recv: str(msg).strip() if msg else None for recv, msg in msg.items()
            }
        else:
            # a single expected string and thus a single receiver (defaults to caller)
            receiver = receiver if receiver else caller
            receiver_mapping[receiver] = str(msg).strip() if msg is not None else None

        unmocked_msg_methods = {}
        for receiver in receiver_mapping:
            # save the old .msg method so we can get it back
            # cleanly  after the test
            unmocked_msg_methods[receiver] = receiver.msg
            # replace normal `.msg` with a mock
            receiver.msg = Mock()

        # Run the methods of the Command. This mimics what happens in the
        # cmdhandler. This will have the mocked .msg be called as part of the
        # execution. Mocks remembers what was sent to them so we will be able
        # to retrieve what was sent later.
        try:
            if cmdobj.at_pre_cmd():
                return
            cmdobj.parse()
            ret = cmdobj.func()

            # handle func's with yield in them (making them generators)
            if isinstance(ret, types.GeneratorType):
                while True:
                    try:
                        inp = inputs.pop() if inputs else None
                        if inp:
                            try:
                                # this mimics a user's reply to a prompt
                                ret.send(inp)
                            except TypeError:
                                next(ret)
                                ret = ret.send(inp)
                        else:
                            # non-input yield, like yield(10). We don't pause
                            # but fire it immediately.
                            next(ret)
                    except StopIteration:
                        break

            cmdobj.at_post_cmd()
        except StopIteration:
            pass
        except InterruptCommand:
            pass

        for inp in inputs:
            # if there are any inputs left, we may have a non-generator
            # input to handle (get_input/ask_yes_no that uses a separate
            # cmdset rather than a yield
            caller.execute_cmd(inp)

        # At this point the mocked .msg methods on each receiver will have
        # stored all calls made to them (that's a basic function of the Mock
        # class). We will not extract them and compare to what we expected to
        # go to each receiver.

        returned_msgs = {}
        for receiver, expected_msg in receiver_mapping.items():
            # get the stored messages from the Mock with Mock.mock_calls.
            stored_msg = [
                args[0] if args and args[0] else kwargs.get("text", to_str(kwargs))
                for name, args, kwargs in receiver.msg.mock_calls
            ]
            # we can return this now, we are done using the mock
            receiver.msg = unmocked_msg_methods[receiver]

            # Get the first element of a tuple if msg received a tuple instead of a string
            stored_msg = [
                str(smsg[0]) if isinstance(smsg, tuple) else str(smsg) for smsg in stored_msg
            ]
            if expected_msg is None:
                # no expected_msg; just build the returned_msgs dict

                returned_msg = "\n".join(str(msg) for msg in stored_msg)
                returned_msgs[receiver] = ansi.parse_ansi(returned_msg, strip_ansi=noansi).strip()
            else:
                # compare messages to expected

                # set our separator for returned messages based on parsing ansi or not
                msg_sep = "|" if noansi else "||"

                # We remove Evmenu decorations since that just makes it harder
                # to write the comparison string. We also strip ansi before this
                # comparison since otherwise it would mess with the regex.
                returned_msg = msg_sep.join(
                    _RE_STRIP_EVMENU.sub("", ansi.parse_ansi(mess, strip_ansi=noansi))
                    for mess in stored_msg
                ).strip()

                # this is the actual test
                if expected_msg == "" and returned_msg or not returned_msg.startswith(expected_msg):
                    # failed the test
                    raise AssertionError(
                        self._ERROR_FORMAT.format(
                            expected_msg=expected_msg, returned_msg=returned_msg
                        )
                    )
                # passed!
                returned_msgs[receiver] = returned_msg

        if len(returned_msgs) == 1:
            return list(returned_msgs.values())[0]
        return returned_msgs


# Base testing classes


@override_settings(**DEFAULT_SETTINGS)
class BaseEvenniaTestCase(TestCase):
    """
    Base test (with no default objects) but with enforced default settings.

    """

    def tearDown(self) -> None:
        super().tearDown()
        flush_cache()


class EvenniaTestCase(TestCase):
    """
    For use with gamedir settings; Just like the normal test case, only for naming consistency.

    Notes:

    -   Inheriting from this class will bypass EvenniaTestMixin, and therefore
        not setup some default objects. This can result in faster tests.

    -   If you do inherit from this class for your unit tests, and have
        overridden the tearDown() method, please also call flush_cache(). Not
        doing so will result in flakey and order-dependent tests due to the
        Django ID cache not being flushed.
    """

    def tearDown(self) -> None:
        super().tearDown()
        flush_cache()


@override_settings(**DEFAULT_SETTINGS)
class BaseEvenniaTest(EvenniaTestMixin, TestCase):
    """
    This class parent has all default objects and uses only default settings.

    """


class EvenniaTest(EvenniaTestMixin, TestCase):
    """
    This test class is intended for inheriting in mygame tests.
    It helps ensure your tests are run with your own objects
    and settings from your game folder.

    """

    account_typeclass = settings.BASE_ACCOUNT_TYPECLASS
    object_typeclass = settings.BASE_OBJECT_TYPECLASS
    character_typeclass = settings.BASE_CHARACTER_TYPECLASS
    exit_typeclass = settings.BASE_EXIT_TYPECLASS
    room_typeclass = settings.BASE_ROOM_TYPECLASS
    script_typeclass = settings.BASE_SCRIPT_TYPECLASS


@patch("evennia.commands.account.COMMAND_DEFAULT_CLASS", MuxCommand)
@patch("evennia.commands.admin.COMMAND_DEFAULT_CLASS", MuxCommand)
@patch("evennia.commands.batchprocess.COMMAND_DEFAULT_CLASS", MuxCommand)
@patch("evennia.commands.building.COMMAND_DEFAULT_CLASS", MuxCommand)
@patch("evennia.commands.comms.COMMAND_DEFAULT_CLASS", MuxCommand)
@patch("evennia.commands.general.COMMAND_DEFAULT_CLASS", MuxCommand)
@patch("evennia.commands.help.COMMAND_DEFAULT_CLASS", MuxCommand)
@patch("evennia.commands.syscommands.COMMAND_DEFAULT_CLASS", MuxCommand)
@patch("evennia.commands.system.COMMAND_DEFAULT_CLASS", MuxCommand)
@patch("evennia.commands.unloggedin.COMMAND_DEFAULT_CLASS", MuxCommand)
@override_settings(**DEFAULT_SETTINGS)
class BaseEvenniaCommandTest(BaseEvenniaTest, EvenniaCommandTestMixin):
    """
    Commands only using the default settings.

    """


class EvenniaCommandTest(EvenniaTest, EvenniaCommandTestMixin):
    """
    Parent class to inherit from - makes tests use your own
    classes and settings in mygame.

    """



"""
Module containing the CallbackHandler for individual objects.
"""

from collections import namedtuple


class CallbackHandler(object):

    """
    The callback handler for a specific object.

    The script that contains all callbacks will be reached through this
    handler.  This handler is therefore a shortcut to be used by
    developers.  This handler (accessible through `obj.callbacks`) is a
    shortcut to manipulating callbacks within this object, getting,
    adding, editing, deleting and calling them.

    """

    script = None

    def __init__(self, obj):
        self.obj = obj

    def all(self):
        """
        Return all callbacks linked to this object.

        Returns:
            All callbacks in a dictionary callback_name: callback}.  The callback
            is returned as a namedtuple to simplify manipulation.

        """
        callbacks = {}
        handler = type(self).script
        if handler:
            dicts = handler.get_callbacks(self.obj)
            for callback_name, in_list in dicts.items():
                new_list = []
                for callback in in_list:
                    callback = self.format_callback(callback)
                    new_list.append(callback)

                if new_list:
                    callbacks[callback_name] = new_list

        return callbacks

    def get(self, callback_name):
        """
        Return the callbacks associated with this name.

        Args:
            callback_name (str): the name of the callback.

        Returns:
            A list of callbacks associated with this object and of this name.

        Note:
            This method returns a list of callback objects (namedtuple
            representations).  If the callback name cannot be found in the
            object's callbacks, return an empty list.

        """
        return self.all().get(callback_name, [])

    def get_variable(self, variable_name):
        """
        Return the variable value or None.

        Args:
            variable_name (str): the name of the variable.

        Returns:
            Either the variable's value or None.

        """
        handler = type(self).script
        if handler:
            return handler.get_variable(variable_name)

        return None

    def add(self, callback_name, code, author=None, valid=False, parameters=""):
        """
        Add a new callback for this object.

        Args:
            callback_name (str): the name of the callback to add.
            code (str): the Python code associated with this callback.
            author (Character or Account, optional): the author of the callback.
            valid (bool, optional): should the callback be connected?
            parameters (str, optional): optional parameters.

        Returns:
            The callback definition that was added or None.

        """
        handler = type(self).script
        if handler:
            return self.format_callback(
                handler.add_callback(
                    self.obj, callback_name, code, author=author, valid=valid, parameters=parameters
                )
            )

    def edit(self, callback_name, number, code, author=None, valid=False):
        """
        Edit an existing callback bound to this object.

        Args:
            callback_name (str): the name of the callback to edit.
            number (int): the callback number to be changed.
            code (str): the Python code associated with this callback.
            author (Character or Account, optional): the author of the callback.
            valid (bool, optional): should the callback be connected?

        Returns:
            The callback definition that was edited or None.

        Raises:
            RuntimeError if the callback is locked.

        """
        handler = type(self).script
        if handler:
            return self.format_callback(
                handler.edit_callback(
                    self.obj, callback_name, number, code, author=author, valid=valid
                )
            )

    def remove(self, callback_name, number):
        """
        Delete the specified callback bound to this object.

        Args:
            callback_name (str): the name of the callback to delete.
            number (int): the number of the callback to delete.

        Raises:
            RuntimeError if the callback is locked.

        """
        handler = type(self).script
        if handler:
            handler.del_callback(self.obj, callback_name, number)

    def call(self, callback_name, *args, **kwargs):
        """
        Call the specified callback(s) bound to this object.

        Args:
            callback_name (str): the callback name to call.
            *args: additional variables for this callback.

        Keyword Args:
            number (int, optional): call just a specific callback.
            parameters (str, optional): call a callback with parameters.
            locals (dict, optional): a locals replacement.

        Returns:
            True to report the callback was called without interruption,
            False otherwise.  If the callbackHandler isn't found, return
            None.

        """
        handler = type(self).script
        if handler:
            return handler.call(self.obj, callback_name, *args, **kwargs)

        return None

    @staticmethod
    def format_callback(callback):
        """
        Return the callback namedtuple to represent the specified callback.

        Args:
            callback (dict): the callback definition.

        The callback given in argument should be a dictionary containing
        the expected fields for a callback (code, author, valid...).

        """
        if "obj" not in callback:
            callback["obj"] = None
        if "name" not in callback:
            callback["name"] = "unknown"
        if "number" not in callback:
            callback["number"] = -1
        if "code" not in callback:
            callback["code"] = ""
        if "author" not in callback:
            callback["author"] = None
        if "valid" not in callback:
            callback["valid"] = False
        if "parameters" not in callback:
            callback["parameters"] = ""
        if "created_on" not in callback:
            callback["created_on"] = None
        if "updated_by" not in callback:
            callback["updated_by"] = None
        if "updated_on" not in callback:
            callback["updated_on"] = None

        return Callback(**callback)


Callback = namedtuple(
    "Callback",
    (
        "obj",
        "name",
        "number",
        "code",
        "author",
        "valid",
        "parameters",
        "created_on",
        "updated_by",
        "updated_on",
    ),
)



"""
Module containing the test cases for the Audit system.

"""

import re

from anything import Anything
from django.test import override_settings
from mock import patch

import evennia
from evennia.utils.test_resources import BaseEvenniaTest

from .server import AuditedServerSession


@override_settings(AUDIT_MASKS=[])
class AuditingTest(BaseEvenniaTest):
    @patch("evennia.server.sessionhandler._ServerSession", AuditedServerSession)
    def setup_session(self):
        """Overrides default one in EvenniaTest"""
        dummysession = AuditedServerSession()
        dummysession.init_session("telnet", ("localhost", "testmode"), evennia.SESSION_HANDLER)
        dummysession.sessid = 1
        evennia.SESSION_HANDLER.portal_connect(
            dummysession.get_sync_data()
        )  # note that this creates a new Session!
        session = evennia.SESSION_HANDLER.session_from_sessid(1)  # the real session
        evennia.SESSION_HANDLER.login(session, self.account, testmode=True)
        self.session = session

    @patch(
        "evennia.contrib.utils.auditing.server.AUDIT_CALLBACK",
        "evennia.contrib.utils.auditing.outputs.to_syslog",
    )
    @patch("evennia.contrib.utils.auditing.server.AUDIT_IN", True)
    @patch("evennia.contrib.utils.auditing.server.AUDIT_OUT", True)
    def test_mask(self):
        """
        Make sure the 'mask' function is properly masking potentially sensitive
        information from strings.
        """
        safe_cmds = (
            "/say hello to my little friend",
            "@ccreate channel = for channeling",
            "@create/drop some stuff",
            "@create rock",
            "@create a pretty shirt : evennia.contrib.game_systems.clothing.Clothing",
            "@charcreate johnnyefhiwuhefwhef",
            'Command "@logout" is not available. Maybe you meant "@color" or "@cboot"?',
            '/me says, "what is the password?"',
            "say the password is plugh",
            # Unfortunately given the syntax, there is no way to discern the
            # latter of these as sensitive
            "@create pretty sunset" "@create johnny password123",
            '{"text": "Command \'do stuff\' is not available. Type "help" for help."}',
        )

        for cmd in safe_cmds:
            self.assertEqual(self.session.mask(cmd), cmd)

        unsafe_cmds = (
            (
                "something - new password set to 'asdfghjk'.",
                "something - new password set to '********'.",
            ),
            (
                "someone has changed your password to 'something'.",
                "someone has changed your password to '*********'.",
            ),
            ("connect johnny password123", "connect johnny ***********"),
            ("concnct johnny password123", "concnct johnny ***********"),
            ("concnct johnnypassword123", "concnct *****************"),
            ('connect "johnny five" "password 123"', 'connect "johnny five" **************'),
            ('connect johnny "password 123"', "connect johnny **************"),
            ("create johnny password123", "create johnny ***********"),
            ("@password password1234 = password2345", "@password ***************************"),
            ("@password password1234 password2345", "@password *************************"),
            ("@passwd password1234 = password2345", "@passwd ***************************"),
            ("@userpassword johnny = password234", "@userpassword johnny = ***********"),
            ("craete johnnypassword123", "craete *****************"),
            (
                "Command 'conncect teddy teddy' is not available. Maybe you meant \"@encode\"?",
                "Command 'conncect ******** ********' is not available. Maybe you meant \"@encode\"?",
            ),
            (
                "{'text': u'Command \\'conncect jsis dfiidf\\' is not available. Type \"help\" for help.'}",
                "{'text': u'Command \\'conncect jsis ********\\' is not available. Type \"help\" for help.'}",
            ),
        )

        for index, (unsafe, safe) in enumerate(unsafe_cmds):
            self.assertEqual(re.sub(" <Masked: .+>", "", self.session.mask(unsafe)).strip(), safe)

        # Make sure scrubbing is not being abused to evade monitoring
        secrets = [
            "say password password password; ive got a secret that i cant explain",
            "whisper johnny = password\n let's lynch the landlord",
            "say connect johnny password1234|the secret life of arabia",
            "@password eval(\"__import__('os').system('clear')\", {'__builtins__':{}})",
        ]
        for secret in secrets:
            self.assertEqual(self.session.mask(secret), secret)

    @patch(
        "evennia.contrib.utils.auditing.server.AUDIT_CALLBACK",
        "evennia.contrib.utils.auditing.outputs.to_syslog",
    )
    @patch("evennia.contrib.utils.auditing.server.AUDIT_IN", True)
    @patch("evennia.contrib.utils.auditing.server.AUDIT_OUT", True)
    def test_audit(self):
        """
        Make sure the 'audit' function is returning a dictionary based on values
        parsed from the Session object.
        """
        log = self.session.audit(src="client", text=[["hello"]])
        obj = {
            k: v for k, v in log.items() if k in ("direction", "protocol", "application", "text")
        }
        self.assertEqual(
            obj,
            {
                "direction": "RCV",
                "protocol": "telnet",
                "application": Anything,  # this will change if running tests from the game dir
                "text": "hello",
            },
        )

        # Make sure OOB data is being recorded
        log = self.session.audit(
            src="client", text="connect johnny password123", prompt="hp=20|st=10|ma=15", pane=2
        )
        self.assertEqual(log["text"], "connect johnny ***********")
        self.assertEqual(log["data"]["prompt"], "hp=20|st=10|ma=15")
        self.assertEqual(log["data"]["pane"], 2)
